#!/usr/bin/env perl

# Supermap: Piecewise monotonic alignment map generator for Shuffle-LAGAN
# Author: Andrey Kislyuk (kislyuk@ocf.berkeley.edu)

package Supermap;
require 5.005;

my $VERSION = (
  '$Id: supermap.pl,v 1.88 2012/06/20 04:20:37 poliakov Exp $'
  =~ /,v\s+(\d+\S+)/o)[0] || "TEST";

# Default constant values
my $overlap_factor = 0.8; # Aligns will be discarded if another align overlaps them by this factor or more in both seqs and has the same orientation
my $max_asym = 10; # Chains will be formed only if the resulting region's lengths differ by at most this factor
my $min_seq_score; # All aligns for sequences with this total score will be discarded. See getMinSeqScore
my $max_expand_len = 30000; # Aligns will be expanded or contracted on both sides on both strands by this amount up to the total length below
my $expand_factor = 4; # When one of an align's sequences is constrained in its expansion by a neighbor/start/end, the other one will be expanded by this times more than the first one
my $max_chainlen = 1500000; # Aligns will not be joined if the total length on either strand exceeds this. Set 0 to disable (no chain length limit)
my $max_job_size = 50000; # Maximum job size, in blat hits, for chunking when running glocal in parallel
my $erode_align = 25; # Amount by which to erode the coords of each align loaded (to avoid overlap problems when chaining)
my ($c1, $c2, $c3, $c4) = (100, 50, 400, 25); # BLAT->CHAOS score conversion parameters
my $default_lagan_dir = "/home/genome/lagan";
my $glocal_name = "glocal";

use Getopt::Long;
use File::Path;
use File::Copy;
use Cwd;
use IPC::Open2;
use IO::Handle;
#use Carp;
use strict;
use warnings;
no warnings "uninitialized";

sub main();
sub init();
sub getSeqSizes($$$);
sub prepareHits();
sub runSLAGAN();
sub reprintInputHits($$$);
sub removeSLAGANOutput();
sub seqBelowMinScore($);
sub alignHashID($);
sub workerRun($$$$);
sub dequeueClustJobs($);
sub get_all_seqs($$);
sub isBLAT($);
sub useIf($$);
sub writeSizes($$);
sub getMinSeqScore($);
sub checkAlignCoords($);
sub expandSeq1($$$);
sub expandSeq2($$$);
sub finalExpand($$);

# array index constants
use constant START1 =>  0; use constant END1   =>  1;
use constant START2 =>  2; use constant END2   =>  3;
use constant SEQ1   =>  4; use constant SEQ2   =>  5;
use constant ORIENT =>  6; use constant ORIGIN =>  7;
use constant SCORE  =>  8; use constant TOTSC  =>  9;
use constant HASHID => 10; use constant FLIPPED=> 11;
use constant CHALO1 => 12; use constant CHAHI1 => 13;
use constant CHALO2 => 14; use constant CHAHI2 => 15;
use constant CHALO1E=> 16; use constant CHAHI1E=> 17;
use constant CHALO2E=> 18; use constant CHAHI2E=> 19;
#use constant PREV1  =>  8; use constant NEXT1  =>  9;
#use constant PREV2  => 10; use constant NEXT2  => 11;
#use constant OSTART1=> 12; use constant OEND1  => 13;
#use constant OSTART2=> 14; use constant OEND2  => 15;

$SIG{'INT'} = $SIG{'QUIT'} = $SIG{'HUP'} = $SIG{'TRAP'} = $SIG{'ABRT'} =
  $SIG{'STOP'} = $SIG{'TERM'} = \&dequeueClustJobs;
$ENV{LC_ALL} = "C";

my $MAX_FILE_SIZE = 1500;
my ($debug, $quiet, $outfile, $proflip, $skip, $no_pid, $input_glob, $input_dir,
  $server, $db, $gen1, $gen2, $gen1sizefile, $gen2sizefile, $write_sizes1,
  $write_sizes2, $score_file, $cfg, $cfg_file, $sizes1, $sizes2, $dbh, $tmp_dir,
  $tmp_prefix, $nodelete, $clust_run_pid, $print_chains, $no_aligntotals,
  $no_clust_run, $num_jobs, $input_is_blat, $force_overwrite, $print_csv,
  $using_GP, $tmp_existed, $print_stats, $lagan_dir, $glocal_out_logfile,
  @input_files, %offsets1, %offsets2, %aligns1, %aligns2, %flipped_aligns,
  $my_cnf, $group, $use_nfs_dir, $submit_dir);

my $supermapexec = $0;
my $mycwd = getcwd();
$supermapexec =~ s/^\./$mycwd/ unless ($supermapexec =~ /^\.\./);
$supermapexec = $mycwd."/".$supermapexec if ($supermapexec =~ /^\.\./);
if (!(-f $supermapexec) && ($ARGV[0] ne "worker")) {
  die "$0: Problem resolving my name, \'$supermapexec\' is not a file";
}
$0 = rindex($0, "/") > -1 ? substr($0, rindex($0, "/")+1) : $0;

if (defined($ENV{"LAGAN_DIR"})) {
  $lagan_dir = $ENV{"LAGAN_DIR"};
} else {
  $lagan_dir = $ENV{"LAGAN_DIR"} = $default_lagan_dir;
}
$lagan_dir =~ /^\// or $lagan_dir = "$mycwd/$lagan_dir";
$lagan_dir =~ s/\/+$//;
$ENV{"LAGAN_DIR"} ne $lagan_dir and $ENV{"LAGAN_DIR"} = $lagan_dir;
# make sure that lagan_dir is mounted on the cluster node
if ($ARGV[0] eq "worker") {
  foreach (1..600) {
    system("/bin/ls '$lagan_dir' >/dev/null 2>&1") or last;
    sleep(3);
  }
}
-d $lagan_dir or die "$0: LAGAN_DIR=$lagan_dir is not a valid directory\n";
push(@INC, $lagan_dir);

my $SLAGAN = "$lagan_dir/$glocal_name";
my $error_file = "./$0.$$.error.log";
my $default_score_file = "$lagan_dir/test.score";
my $default_outfile = "$0.out";
# The directory where workers store their intermediate files (two workers 
# should not use the same directory)
my $worker_tmp_dir = "/tmp/$0.$$.worker";

my $usage = "
-infile=file \t Name of input file containing all hits for the two genomes
-outfile=file \t Output filename (default: $default_outfile)
-gen1=id \t First genome ID (must exist in the GPDB)
-gen2=id \t Second genome ID (must exist in the GPDB)
-sizes1=file \t File with sequence sizes for first genome
-sizes2=file \t File with sequence sizes for second genome
-bacteria \t Rearrange circular DNA to find a better alignment map
-server=hostname GPDB server (default: lemur)
-db=dbname \t GPDB name (default: GP)
-config=file \t GPDB config file (default: ~/.gprc)
-score=file \t Score file for SLAGAN (default: $default_score_file)
-glocal_out=file \t Save intermediate GLOCAL alignment hits to this file
-no_clust_run \t Run CPU/memory intensive jobs locally, not on the GP cluster
-tmp_dir=dir \t Working directory (default: /tmp/$0.pid)
-f \t\t Overwrite output file without prompting if it exists
-v \t\t Verbose mode
-q \t\t Quiet mode
-k \t\t Keep all temporary files
-expand_length=N Maximum length by which to expand alignments (default: $max_expand_len)
-max_asym=N \t Maximum asymmetry factor between sequences for chaining
\t\t (default: $max_asym)
-max_length=N \t Maximum length for any alignment chain in either strand
\t\t (default: $max_chainlen)
-min_seq_score=N Sequences with total align score below this threshold will be
\t\t discarded (default: U penalty in SLAGAN score file)
-max_job_size=N  Threshold, in hits, for splitting workload into separate jobs
\t\t for clust_run (default: $max_job_size)
-c1, c2, c3, c4=N: Score factors for BLAT->CHAOS conversion
\t\t (default: $c1, $c2, $c3, $c4)

Options may be abbreviated.
Input file format is BLAT or CHAOS. Sequence names should not contain spaces.
Alignments with negative scores are discarded.
Sequence size file format, one sequence per line: seq_name seq_size
";

exit(main());

# ___ Subroutines _______________

sub main() {
  my (%sort1_pos, %sort2_pos);

  if ($ARGV[0] eq "worker") {
    # Running SLAGAN in distributed mode
    workerRun($ARGV[1], $ARGV[2], $ARGV[3], $ARGV[4]);
    return 0;
  }
  init();

  print("$0: Retrieving sequence info...\n") unless $quiet;
  $sizes1 = getSeqSizes($dbh, $gen1, $gen1sizefile);
  (writeSizes($sizes1, $write_sizes1), return 0) if defined $write_sizes1;
  $sizes2 = getSeqSizes($dbh, $gen2, $gen2sizefile);
  (writeSizes($sizes2, $write_sizes2), return 0) if defined $write_sizes2;
  $using_GP && ($gen1 || $gen2) and $dbh->disconnect();

  die("$0: No sequence size data found. Stopped")
    if (keys(%$sizes1) < 1 or keys(%$sizes2) < 1);
  die("$0: Flip mode is only applicable for two single-sequence organisms. " .
      "Stopped")
    if ($proflip and not (keys(%$sizes1) == 1 and keys(%$sizes2) == 1));

  # Sort and separate the alignments, run SLAGAN on them
  prepareHits();
  runSLAGAN();

  # Chain SLAGAN alignments into supermonotonic chain and
  # save the intermediate results
  processResults(\%sort1_pos, \%sort2_pos);

  # Load the results back and expand regions, then print them
  postProcessRegions(\%sort1_pos, \%sort2_pos);

  $quiet or print "$0: Output written to $outfile\n";
  $nodelete && !$quiet and print "$0: Intermediate files kept in $tmp_dir\n";
  $tmp_existed || $nodelete or rmtree($tmp_dir);
  
  return 0;
}


# Startup tasks
sub init() {
  # Berkeley Genome Pipeline functionality is used if corresponding 
  # Perl modules are found in @INC
  foreach my $dir (@INC) {
    $using_GP = 1 if -f "$dir/GPDBI.pm" and -f "$dir/GPutils.pm";
  }

  useIf $using_GP, "GPDBI";
  useIf $using_GP, "GPutils";
  useIf 1, "Utils";

  die("$0: GetOptions failed to retrieve options. Check the input options. " .
    "Usage:$usage") unless
  GetOptions(
    "config=s"    => \$cfg_file,
    "server=s"    => \$server,
    "db=s"      => \$db,
    "my_cnf=s" => \$my_cnf,
    "group=s" => \$group,
    "gen1=s"    => \$gen1,
    "gen2=s"    => \$gen2,
    "sizes1=s"    => \$gen1sizefile,
    "sizes2=s"    => \$gen2sizefile,
    "blatfile=s"  => \$input_glob,
    "infile=s"    => \$input_glob,
    "outfile=s"    => \$outfile,
    "glocal_out=s"  => \$glocal_out_logfile,
    "bacteria"    => \$proflip,
    "tmp_dir=s"    => \$tmp_dir,
    "skip"      => \$skip,
    "no_pid"    => \$no_pid,
    "no_clust_run"  => \$no_clust_run,
    "print_chains"  => \$print_chains,
    "print_stats"  => \$print_stats,
    "no_aligntotals"=> \$no_aligntotals,
    "print_csv"    => \$print_csv,
    "max_job_size"  => \$max_job_size,
    "max_length=i"  => \$max_chainlen,
    "expand_length=i"=>\$max_expand_len,
    "min_seq_score=i"=>\$min_seq_score,
    "max_asym=i"  => \$max_asym,
    "overlap_factor"=> \$overlap_factor,
    "score=s"    => \$score_file,
    "c1=i"      => \$c1,
    "c2=i"      => \$c2,
    "c3=i"      => \$c3,
    "c4=i"      => \$c4,
    "write_sizes1=s"=> \$write_sizes1,
    "write_sizes2=s"=> \$write_sizes2,
    "keep"      => \$nodelete,
    "f"        => \$force_overwrite,
    "v"        => \$debug,
    "q"        => \$quiet
  );

  undef $quiet if $debug;
  my $uname = getpwuid($>);
  print("$0: Version $VERSION started " . localtime() . " by $uname\n")
    unless $quiet;
  $tmp_prefix = $0.($no_pid ? "" : ".".$$);

  unless ($no_clust_run) {
    $no_clust_run = `which clust_run 2> /dev/null`;
    $no_clust_run = not $no_clust_run;
    print("$0: clust_run not found - cluster operation disabled\n")
      if ($no_clust_run && !$quiet);
  }

  if ($tmp_dir) {
    $tmp_dir = Utils::Get_Abs_Path($tmp_dir);
    if (-d $tmp_dir) {
      $tmp_existed = 1;
    } else {
      mkdir $tmp_dir;
    }
  } else {
    $tmp_dir = "/tmp/$tmp_prefix";
    mkdir $tmp_dir;
  }
  -w $tmp_dir or
    die "$0: No write permissions in working directory $tmp_dir. Stopped";
  $tmp_dir =~ s/\/+$//;
  die("$0: Genome IDs or size files not specified. Usage:".$usage)
    unless ($gen1 or $gen1sizefile) and ($gen2 or $gen2sizefile);
  die("$0: '-gen' options are invalid because GPDB is not available. " .
      "Use '-sizes'. Stopped")
    if (($gen1 or $gen2) and not $using_GP);
  die("$0: Sequence size file $gen1sizefile not found. Stopped")
    unless -f $gen1sizefile or $gen1;
  die("$0: Sequence size file $gen2sizefile not found. Stopped")
    unless -f $gen2sizefile or $gen2;
  die("$0: Maximum job size too small, must exceed 10000 hits. Stopped")
    if $max_job_size < 10000;
  die("$0: Overlap factor must be between 0 and 1. Stopped")
    if $overlap_factor < 0 or $overlap_factor > 1;
  $quiet or
    print "$0: using SLAGAN score file " .
      ($score_file || "$default_score_file (default)") . "\n";
  $quiet or
    print "$0: using output file " .
      ($outfile || "$default_outfile (default)") . "\n";

  # Check input file or glob
  $input_dir = "";
  if (defined $input_glob) {
    if ($input_glob =~ /\//) {
      ($input_dir, $input_glob) = ($input_glob =~ /^(.*)\/([^\/]+)$/o);
    }
    $input_glob .= "\$" unless $input_glob =~ /\$$/;
    $input_glob = "^".$input_glob unless $input_glob =~ /^\^/;
    @input_files = Utils::safe_glob($input_glob, $input_dir);
  } elsif (@ARGV > 0) {
    foreach my $file (@ARGV) {
      if ($file =~ /\//) {
        ($input_dir, $file) = ($file =~ /^(.*)\/([^\/]+)$/o); # BUG
      }
      push @input_files, $file;
    }
  } else { # TODO: split stdin for >2GB input
    open(FH, "> $tmp_dir/$tmp_prefix.in");
    print FH while <STDIN>;
    close FH;
    push @input_files, "$tmp_prefix.in";
    $input_dir = $tmp_dir;
  }
  $input_dir =~ /^\//o or $input_dir = "$mycwd/$input_dir";
  $input_dir =~ s/\/+$//;
  @input_files > 0 or
    die "$0: No input files matching \"$input_dir/$input_glob\" found. Stopped";
  $debug and print "$0: " . scalar(@input_files) . " input file(s)\n";

  # Check output file
  $outfile = $default_outfile unless $outfile;
  if (-f $outfile and not $force_overwrite and -t STDERR) {
    print STDERR "$0: $outfile exists. Overwrite? (y/N, '-f' to force) ";
    my $overwrite = <STDIN>; chomp $overwrite;
    (print("Move \"$outfile\" or use option '-f'.\n"), exit(1))
      unless ($overwrite eq "Y" or $overwrite eq "y" or $overwrite eq "yes");
  }
  open(FH, "> $outfile") or die("$0: Cannot open $outfile for writing: $!");
  close FH;

  # Check SLAGAN score file
  $score_file = $default_score_file unless $score_file;
  $score_file = Get_Abs_Path($score_file);
  $max_expand_len += $erode_align;
  die("$0: max_length cannot be less than 0. Stopped") if $max_chainlen < 0;
  $max_chainlen = 1000000000 if $max_chainlen == 0;
  $max_chainlen -= 2*$max_expand_len;
  # SLAGAN output for a given sequence will be discarded
  # if the total score for the sequence is below this threshold.
  # Default value is the SLAGAN unrelated gap penalty.
  $min_seq_score = getMinSeqScore($score_file) unless defined $min_seq_score;

  if ($using_GP) {
    my ($attr);
    $GPutils::Error = "";
    ($cfg_file, $server, $db, $my_cnf, $group, $attr, undef, $cfg) =
      get_gp_config($cfg_file, $server, $db, $my_cnf, $group) or
      die "$GPutils::Error";
    add2path($cfg->Get_Val("Directories", "bin"));
    !$no_clust_run and $use_nfs_dir = $cfg->Get_Val("cluster", "use_nfs_dir");
    if ($gen1 || $gen2) { # Connect to GPDB
      $dbh = GPDBI->connect($server, 0, $db, undef, $my_cnf, $group, undef,
        $attr);
    }
  }
}


# Load sequence names and sizes either from GPDB or from file

sub getSeqSizes($$$) {
  my ($dbh, $dataset, $gen_size_file) = @_;

  if ($dataset) {
    return get_all_seqs($dbh, $dataset);
  } else {
    my %sizes;
    local (*FH);
    open(FH, "< $gen_size_file")
      or die("$0: Could not open file $gen_size_file for reading: $!");
    while (<FH>) {
      chomp;
      my ($seq, $size) = split;
      die("$0: Invalid format in file $gen_size_file") unless $seq and $size;
      $sizes{$seq} = $size;
    }
    close FH;
    return \%sizes;
  }
}


# Convert BLAT to CHAOS if necessary
# Flip hits on circular sequence if necessary

sub prepareHits() {
  my ($cur_align);
  local (*FH, *OUT1);

  $quiet or print "$0: Preparing files...\n";
  isBLAT("$input_dir/$input_files[0]") and $input_is_blat = 1;

  if ($input_is_blat) {
    foreach my $file (@input_files) {
      system('awk \'{$13=($13+$15)?$13:1; print $1,$2,$3";",$5,$6,$7"; '.
        'score = "' . $c1 . '*$8-' . $c2 . '*$9-' . $c3 . '*($12+$14)-' . $c4 .
        '*log($13+$15),"("$4")"}\''.
        "< '$input_dir/$file' > '$tmp_dir/$file.chaos'");
    }
  } else {
    foreach my $file (@input_files) {
      system("ln -s '$input_dir/$file' '$tmp_dir/$file.chaos'");
    }
  }

  if ($proflip) {
    open(FH, "< $tmp_dir/$input_files[0].chaos")
      or die("$0: Could not open file $tmp_dir/$input_files[0].chaos " .
        "for reading: $!");
    open(OUT1, "> $tmp_dir/$input_files[0].flipped.chaos")
      or die("$0: Could not open file $tmp_dir/$input_files[0].flipped.chaos " .
        "for writing: $!");

    my (@seq1s, @seq1e, @seq2s, @seq2e, @scores, @orientations, @seqn1, @seqn2,
      $seq1center, $seq2center, $seq1median, $seq2median);
    my $i = 0;
    my $re = '^\s*.*\s(\d+)\s(\d+);\s.*\s(\d+)\s(\d+);\sscore\s=' .
      '\s([-+e\d\.]+)\s\(([-+]+)\)';
    while (<FH>) {
      /$re/o;
      ($seq1s[$i], $seq1e[$i], $seq2s[$i], $seq2e[$i], $scores[$i],
        $orientations[$i]) = ($1, $2, $3, $4, $5, $6);
      if ($seq1s[$i] > $seq1e[$i]) {
        my $j = $seq1s[$i]; $seq1s[$i] = $seq1e[$i]; $seq1e[$i] = $j;
      }
      if ($seq2s[$i] > $seq2e[$i]) {
        my $j = $seq2s[$i]; $seq2s[$i] = $seq2e[$i]; $seq2e[$i] = $j;
      }
      $i++;
    }

    # For each interval pair,
    # if the seq1 interval median is greater than seq1 median,
    # and the corresponding interval median in seq2 is less than seq2 median,
    # OR if the seq1 interval median is less than seq1 median,
    # and the corresponding interval median in seq2 is greater
    # than seq2 median,
    # set start of interval in seq1 to 2CoM1 - previous end of interval
    # set end of interval in seq1 to 2CoM1 - previous start of interval
    # flip the orientation (+/-)
    $seq1center = $$sizes1{(keys(%$sizes1))[0]} / 2;
    $seq2center = $$sizes2{(keys(%$sizes2))[0]} / 2;
    my $flip_counter = 0;
    foreach $i (0..$#seq1s) {
      $seq1median = ($seq1s[$i] + $seq1e[$i]) / 2;
      $seq2median = ($seq2s[$i] + $seq2e[$i]) / 2;
      if (($seq1median > $seq1center and $seq2median < $seq2center)
        or ($seq1median < $seq1center and $seq2median > $seq2center)) {
        my $j = $seq2s[$i];
        $seq2s[$i] = (2 * $seq2center) - $seq2e[$i];
        $seq2e[$i] = (2 * $seq2center) - $j;
        $orientations[$i] = ($orientations[$i] eq "+") ? "-" : "+";
        $cur_align = [];
        $$cur_align[START1] = $seq1s[$i];
	$$cur_align[START2] = $seq2s[$i];
        $$cur_align[END1] = $seq1e[$i];
	$$cur_align[END2] = $seq2e[$i];
        $$cur_align[SCORE] = $scores[$i];
	$$cur_align[ORIENT] = $orientations[$i];
        $$cur_align[SEQ1] = (keys(%$sizes1))[0];
	$$cur_align[SEQ2] = (keys(%$sizes2))[0];
        $$cur_align[START1] += $erode_align;
	$$cur_align[END1] -= $erode_align;
        $$cur_align[START2] += $erode_align;
	$$cur_align[END2] -= $erode_align;
        $flipped_aligns{alignHashID($cur_align)} = $cur_align;
        $flip_counter++;
      }
      print OUT1 "seq1 $seq1s[$i] $seq1e[$i]; seq2 $seq2s[$i] $seq2e[$i]; " .
        "score = $scores[$i] ($orientations[$i])\n";
    }
    close FH; close OUT1;
    print "$0: Single-sequence flip mode: " . ($flip_counter + 0) .
      " hits flipped\n" if $debug;
  }
}


# Load all hits into a hash table, then write the hits for each sequence
# into a file
# Run SLAGAN on each of these files, via worker instances either
# on the cluster or sequentially

sub runSLAGAN() {
  my ($clust_run_invoke, $sort_pid2, $sort_pid3,
    $one_seq_mode, $curlen1, $curlen2, $nextlen1,
    $nextlen2, $overlap1, $overlap2, $dump_count);
  my $re = '^\s*(.*)\s(\d+)\s(\d+);\s(.*)\s(\d+)\s(\d+);\sscore\s=\s' .
    '([-+e\d\.]+)\s\(([-+]+)\)';
  local (*RH2, *WH2, *RH3, *WH3, *DUPES, *FH);

  $submit_dir = $tmp_dir;
  if ($use_nfs_dir && is_local_path($tmp_dir)) {
    my $sdir = "submit_$0.$$";
    mkdir($sdir, 0777) or die;
    $submit_dir = Get_Abs_Path($sdir);
  }
  $debug and print "$0: Sorting input hits...\n";
  $debug and open(DUPES, "> supermap.duplicates");
  (keys(%$sizes1) == 1) && (keys(%$sizes2) == 1) and $one_seq_mode = 1;
  $sort_pid2 = open2(\*RH2, \*WH2, "sort -k 1,1 -k 9,9nr | " .
    "filt_chaos -filter | sort -k 1,1 -k 2,2n"); # gen1base
  $sort_pid3 = open2(\*RH3, \*WH3, "sort -k 4,4 -k 9,9nr | " .
    "filt_chaos -filter -rev | sort -k 4,4 -k 5,5n"); # gen2base
  foreach my $file (@input_files) {
    $Utils::Error = "";
    my $inf = open_file("$tmp_dir/$file" . ($proflip ? ".flipped" : "") .
      ".chaos", "<") or
      die $Utils::Error;
    # Scan input, check if start2, end2 are ascending for sorting,
    # erode alignments
    while (<$inf>) {
      my (@cur_align);
      (@cur_align[SEQ1,START1,END1,SEQ2,START2,END2,SCORE,ORIENT]) =
	($_ =~ /$re/o);
      $cur_align[ORIENT] or next;
      $cur_align[SCORE] <= 0 and next;
      if ($one_seq_mode) {
        $cur_align[SEQ1] = (keys(%$sizes1))[0];
        $cur_align[SEQ2] = (keys(%$sizes2))[0];
      }
      checkAlignCoords(\@cur_align);
      if (!(($cur_align[END1] - $cur_align[START1]) <= ($erode_align * 2) ||
          ($cur_align[END2] - $cur_align[START2]) <= ($erode_align * 2))) {
        $cur_align[START1] += $erode_align;
        $cur_align[END1] -= $erode_align;
        $cur_align[START2] += $erode_align;
        $cur_align[END2] -= $erode_align;
      }
      my $line = "$cur_align[SEQ1] $cur_align[START1] " .
        "$cur_align[END1]; $cur_align[SEQ2] $cur_align[START2] " .
	 "$cur_align[END2]; score = $cur_align[SCORE] " .
	 "($cur_align[ORIENT])\n";
      print WH2 $line;
      print WH3 $line;
    }
    close($inf);
  }
  close(WH2);
  close(WH3);
  $num_jobs = reprintInputHits(1, 1, \*RH2);
  close(RH2);
  waitpid($sort_pid2, 0);
  $num_jobs = reprintInputHits(2, $num_jobs, \*RH3);
  $num_jobs--;
  close(RH3);
  waitpid($sort_pid3, 0);
  defined(fileno(DUPES)) and close(DUPES);
  system("/bin/cp -f -p '$score_file' '$submit_dir/score_file'");
  open(FH, "> $tmp_dir/CLUSTER_JOB_PARAMS") or die;
  foreach my $i (1..$num_jobs) {
    print FH "worker JOB$i.tar score_file '$SLAGAN'" . ($debug ? " -v" : "");
    $no_clust_run or
      print FH " << JOB$i.tar,score_file > CLUSTER_JOB_MESSAGES.$i " .
        ">> CLUSTER_JOB_ERRMSG.$i";
    print FH "\n";
  }
  close(FH);
  if ($no_clust_run) {
    open(FH, "< $tmp_dir/CLUSTER_JOB_PARAMS") or die;
    $quiet or 
      print "$0: Running $num_jobs SLAGAN jobs locally...\n";
    while (<FH>) {
      chomp;
      $debug and print("Job $.: \"$0 $_\"\n");
      system("cd '$tmp_dir'; '$supermapexec' $_");
    }
    close FH;
  } else {
    $clust_run_invoke = "clust_run -program='$supermapexec' " .
      "-parameters='$tmp_dir/CLUSTER_JOB_PARAMS' -init_dir='$submit_dir' " .
      "-env LAGAN_DIR -class 1 -type 1 -wait";
    $quiet or
      print "$0: Running $num_jobs" .
        " distributed SLAGAN jobs with clust_run...\n";
    $debug and print "$0: \"$clust_run_invoke\"\n";

    if ($clust_run_pid = fork()) { # I am the parent
      waitpid($clust_run_pid, 0);
    } elsif (!defined($clust_run_pid)) {
      die("$0: Could not fork");
    } else { # I am the child
      exec($clust_run_invoke) or
        die("$0: Could not exec \"$clust_run_invoke\"");
    }
    undef $clust_run_pid;
  }
  foreach my $i (1..$num_jobs) {
    -s "$submit_dir/CLUSTER_JOB_ERRMSG.$i" and
      die "$0: job $i: SLAGAN crashed " .
        "(details in '$submit_dir/CLUSTER_JOB_ERRMSG.$i')\n";
    $nodelete or
      unlink("$submit_dir/CLUSTER_JOB_ERRMSG.$i", "$submit_dir/JOB$i.tar",
        "$submit_dir/CLUSTER_JOB_MESSAGES.$i");
  }

#  unlink "$tmp_dir/$input_glob.chaos" unless $nodelete;
  $nodelete or unlink("$tmp_dir/CLUSTER_JOB_PARAMS", "$submit_dir/score_file");
  
  if (!$nodelete) {
    foreach my $file (@input_files) {
      unlink("$tmp_dir/$file.chaos");
    }
  }
}

sub reprintInputHit {
  my ($base_gen, $align, $FH) = @_;

  print $FH (($base_gen == 1) ?
    ("$$align[SEQ1] $$align[START1] $$align[END1]; $$align[SEQ2] " .
      (($$align[ORIENT] eq "+") ?
        "$$align[START2] $$align[END2]" :
        "$$align[END2] $$align[START2]")) :
    ("$$align[SEQ2] $$align[START2] $$align[END2]; $$align[SEQ1] " .
      (($$align[ORIENT] eq "+") ?
        "$$align[START1] $$align[END1]" :
        "$$align[END1] $$align[START1]"))) .
    "; score = $$align[SCORE] ($$align[ORIENT])\n";
}

sub writeJobFile {
  my ($job_id, $seq_list) = @_;
  local *LIST;

  open(LIST, "| cd '$tmp_dir'; xargs tar -r " .
    "-f '$submit_dir/JOB$job_id.tar'") or
    die "$0: Cannot start xargs (job $job_id): $!. Stopped";
  foreach my $file (keys(%$seq_list)) {
    $file =~ /\/([^\/]+)$/;
    print LIST "$1 ";
  }
  close LIST or
    die "$0: " . ($! ?
      "Error closing xargs pipe (job $job_id): $!" :
      "Exit status $? from xargs (job $job_id)") . ". Stopped";
  foreach my $file (keys(%$seq_list)) {
    $nodelete or unlink($file);
  }
}

# Separate input into files based on sequence name and reverse order
# in gen2base hits

sub reprintInputHits($$$) {
  my ($base_gen, $job_id, $RH) = @_;
  my ($line_count, $prev_seq, %cur_seq_list, %pruned_sizes);
  my $re = '^\s*(.*)\s(\d+)\s(\d+);\s(.*)\s(\d+)\s(\d+);\sscore\s=\s' .
    '([-+e\d\.]+)\s\(([-+]+)\)';
  local (*OUT, *LIST);

  $debug and print "$0: Reprinting hits (base genome $base_gen)...";
  $line_count = 0;
  while (<$RH>) {
    my (@cur_align);
    (@cur_align[SEQ1,START1,END1,SEQ2,START2,END2,SCORE,ORIENT]) =
      ($_ =~ /$re/o);
    my $cur_seq = $cur_align[($base_gen == 1) ? SEQ1 : SEQ2];
    if (!defined($prev_seq) || ($cur_seq ne $prev_seq)) {
      $pruned_sizes{$cur_seq} = (($base_gen == 1) ?
        $$sizes1{$cur_seq} : $$sizes2{$cur_seq});
      $debug and print " $cur_seq";
      defined(fileno(OUT)) and close(OUT);
      my $outf = "$tmp_dir/$input_files[0].gen${base_gen}base.$cur_seq.chaos";
      open(OUT, "> $outf") or
        die("$0: Could not open file $outf for writing: $!");
      if ($line_count > $max_job_size) {
        writeJobFile($job_id, \%cur_seq_list);
        undef %cur_seq_list;
	$line_count = 0;
	$job_id++;
      }
      $cur_seq_list{$outf} = 1;
    }
    reprintInputHit($base_gen, \@cur_align, \*OUT);
    $prev_seq = $cur_seq;
    $line_count++;
  }
  writeJobFile($job_id, \%cur_seq_list);
  $job_id++;
  defined(fileno(OUT)) and close(OUT);
  $debug and print "\n";
  if ($base_gen == 1) {
    $sizes1 = \%pruned_sizes;
  } else {
    $sizes2 = \%pruned_sizes;
  }
  return $job_id;
}


sub seqBelowMinScore($) {
  my ($line) = @_;
  my $re = '^\s*\(\d+\s\d+\)=\(\d+\s\d+\)\s+([-\.\d]+)\s[-+]+\s' .
    '\[([-\.\d]+)\]\s*s1:.*\s*s2:.*$';

  $line =~ /$re/o;
  die("$0: Unable to extract score values from SLAGAN output:\n$line")
    if not defined $2;
  return ($2 < $min_seq_score);
}

sub processResults {
  my ($sort1_pos, $sort2_pos) = @_;
  my ($prev_seq, $discard, $out, $cur_outf, $dropped_seqs, $sort_pid,
    $sort_pid2, %glocal_out_pos, %hashes_pos);
  local (*RH, *WH, *hashesDM_RH, *hashesDM_WH, *GLOCAL_OUT_LOG, *RH_DM, *WH_DM);
  my $re = '^\s*\(\s*(\d+)\s+(\d+)\s*\)\s*=\s*\(\s*(\d+)\s+(\d+)\s*\)(.+)' .
    '\ss1:\s*(\S+)\s+s2:\s*(\S+)$';

  print "$0: Loading SLAGAN output...\n" unless $quiet;
  open(GLOCAL_OUT_LOG, "> $glocal_out_logfile") if $glocal_out_logfile;
  # Sort gen2base aligns on seq1, then seq2, then start2,
  # then print them to an indexed file
  # This file will be loaded on demand when scanning gen1base aligns
  # (chainBase1Hits())
  # input is base 2, key is 9 because a space is expected between s2:
  # and seq2name
  $sort_pid = open2(\*RH, \*WH, "sort -k 9,9 -k 7,7 -k 1.2,1n");
  foreach my $file (safe_glob("^$input_files[0]" . '\.gen2\.JOB\d+\.results$',
      $submit_dir)) {
    my $inf = "$submit_dir/$file";
    my $in = open_file($inf, "<") or die $Utils::Error;
    while (<$in>) {
      my ($cur_seq) = ($_ =~ /\ss1:(\S+)/o);
      defined($cur_seq) or next;
      if (!defined($prev_seq) || ($cur_seq ne $prev_seq)) {
        $discard = seqBelowMinScore($_);
        if ($discard) {
          $debug and
            print "$0: Discarding sequence $cur_seq in genome 2 - " .
	      "score too low (<$min_seq_score)\n";
        }
	$prev_seq = $cur_seq;
      }
      $discard or print WH $_;
    }
    close($in);
  }
  close(WH) or die("$0: Error executing sort");
  undef($prev_seq);
  while (<RH>) {
    my ($cur_seq) = ($_ =~ /\ss2:\s*(\S+)\s*$/o);
    defined($cur_seq) or next;
    if (!defined($prev_seq) || ($cur_seq ne $prev_seq)) {
      my ($pos);
      ($cur_outf, $pos, $out) =
        get_index("$input_files[0].gen2.sorted-gen1.glocal-out", $cur_outf);
      $glocal_out_pos{$cur_seq} = [$cur_outf, $pos];
      $prev_seq = $cur_seq;
    }
    print $out $_;
  }
  defined($out) && defined(fileno($out)) and close($out);
  close(RH);
  waitpid($sort_pid, 0);
  # Sort gen1base aligns on seq1, then start1
  # input is base 1
  $sort_pid = open2(\*RH, \*WH, "sort -k 7,7 -k 1.2,1n");
  undef($prev_seq);
  foreach my $file (safe_glob("^$input_files[0]" . '\.gen1\.JOB\d+\.results$',
      $submit_dir)) {
    my $inf = "$submit_dir/$file";
    my $in = open_file($inf, "<") or die $Utils::Error;
    while (<$in>) {
      my ($cur_seq) = ($_ =~ /\ss1:(\S+)/o);
      defined($cur_seq) or next;
      if (!defined($prev_seq) || ($cur_seq ne $prev_seq)) {
        $discard = seqBelowMinScore($_);
        if ($discard) {
          $dropped_seqs++;
          $debug and
            print "$0: Discarding sequence $cur_seq in genome 1 - " .
	      "score too low (<$min_seq_score)\n";
        }
	$prev_seq = $cur_seq;
      }
      $discard and next;
      print WH $_;
      $glocal_out_logfile and print GLOCAL_OUT_LOG $_;
    }
    close($in);
    $nodelete or unlink($inf);
  }
  close(WH) or die("$0: Error executing sort");
  # Feed the gen1base aligns to the 2M/1M1 chain scanner (chainBase1Hits())
  # The hashesDM handle is used to write 2M aligns' hashes to be sorted
  # in seq2 order
  $quiet or print "$0: Generating supermonotonic map...\n";
  $sort_pid2 = open2(\*hashesDM_RH, \*hashesDM_WH, "sort -k 2,2");
  chainBase1Hits(*RH, *hashesDM_WH, \%glocal_out_pos, $sort1_pos);
  close(RH);
  waitpid($sort_pid, 0);
  close(hashesDM_WH) or die("$0: Error executing sort");
  undef(%glocal_out_pos);
  $nodelete or rm_db_files("$input_files[0].gen2.sorted-gen1.glocal-out");
  # Print sorted 2M aligns' hashes to an indexed file
  $prev_seq = $cur_outf = undef;
  while(<hashesDM_RH>) {
    my ($cur_seq) = ($_ =~ /^\S+\s(\S+)\s\S+$/o);
    defined($cur_seq) or next;
    if (!defined($prev_seq) || ($cur_seq ne $prev_seq)) {
      my ($pos);
      ($cur_outf, $pos, $out) =
        get_index("$input_files[0].gen2.hashesDM", $cur_outf);
      $hashes_pos{$cur_seq} = [$cur_outf, $pos];
      $prev_seq = $cur_seq;
    }
    print $out $_;
  }
  defined($out) && defined(fileno($out)) and close($out);
  close(hashesDM_RH);
  waitpid($sort_pid2, 0);
  # sort raw DM chains on seq2, start2
  $sort_pid2 = open2(\*RH_DM, \*WH_DM, "sort -k 8,8 -k 5,5n");
  foreach my $file (safe_glob("^$input_files[0]" . '\.rawDM\.\w{10}$',
      $tmp_dir)) {
    copy("$tmp_dir/$file", \*WH_DM);
    $nodelete or unlink("$tmp_dir/$file");
  }
  close(WH_DM) or die("$0: Error executing sort");
  # Sort gen2base aligns on seq2, then start2
  # input is base 2
  $sort_pid = open2(\*RH, \*WH, "sort -k 7,7 -k 1.2,1n");
  undef($prev_seq);
  foreach my $file (safe_glob("^$input_files[0]" . '\.gen2\.JOB\d+\.results$',
      $submit_dir)) {
    my $inf = "$submit_dir/$file";
    my $in = open_file($inf, "<") or die $Utils::Error;
    while (<$in>) {
      my ($cur_seq) = ($_ =~ /\ss1:\s*(\S+)/o);
      defined($cur_seq) or next;
      if (!defined($prev_seq) || ($cur_seq ne $prev_seq)) {
        $discard = seqBelowMinScore($_);
        if ($discard) {
          $dropped_seqs++;
          $debug and
            print "$0: Discarding sequence $cur_seq in genome 2 - " .
	      "score too low (<$min_seq_score)\n";
        }
	$prev_seq = $cur_seq;
      }
      $discard and next;
      print WH $_;
      if ($glocal_out_logfile) {
        my ($start2, $end2, $start1, $end1, $other, $seq_name2, $seq_name1) =
	  ($_ =~ /$re/o);
        ($start1 > $end1) and
	  ($start1, $end1, $start2, $end2) = ($end1, $start1, $end2, $start2);
        print GLOCAL_OUT_LOG "($start1 $end1)=($start2 $end2)$other " .
	  "s1:$seq_name1 s2: $seq_name2\n";
      }
    }
    close($in);
    $nodelete or unlink($inf);
  }
  close(WH) or die("$0: Error executing sort");
  !$nodelete && ($submit_dir ne $tmp_dir) and rmtree($submit_dir);
  # Feed the gen2base aligns and raw DM chains 
  # to the 1M2 chain scanner (chainBase2Hits())
  chainBase2Hits(*RH, \%hashes_pos, $sort2_pos, *RH_DM);
  close(RH);
  waitpid($sort_pid, 0);
  undef(%hashes_pos);
  $nodelete or rm_db_files("$input_files[0].gen2.hashesDM");
  close(RH_DM);
  waitpid($sort_pid2, 0);
  defined(fileno(GLOCAL_OUT_LOG)) and close GLOCAL_OUT_LOG;
  print STDERR "$0: Warning: Alignments for $dropped_seqs sequences " .
    "discarded due to total score below cutoff ($min_seq_score)\n"
    if $dropped_seqs and not $quiet;
}

{
my (%files);

sub get_index {
  my ($file, $suffix) = @_;
  my ($out, $pos);
  
  if (!defined($suffix)) {
    $suffix = 1;
    $out = open_file("$tmp_dir/$file.$suffix", ">") or die $Utils::Error;
    $files{$file} ||= {};
    $files{$file}{$suffix} = $out;
    $pos = 0;
  } else {
    $out = $files{$file}{$suffix};
    $pos = tell($out);
  }
  if ($pos > ($MAX_FILE_SIZE * 1000000)) {
    close($out);
    $suffix++;
    $out = open_file("$tmp_dir/$file.$suffix", ">") or die $Utils::Error;
    $files{$file}{$suffix} = $out;
    $pos = 0;
  }
  return ($suffix, $pos, $out);
}

sub open_db_file {
  my ($file, $suffix, $pos) = @_;
  my ($in);
  
  defined($suffix) or return undef;
  $in = $files{$file}{$suffix};
  if (!defined($in) || !defined(fileno($in))) {
    $in = open_file("$tmp_dir/$file.$suffix", "<") or die $Utils::Error;
    $files{$file}{$suffix} = $in;
  }
  seek($in, $pos, 0) or die "$0: cannot set file position";
  return $in;
}

sub rm_db_files {
  my ($file) = @_;

  foreach my $suffix (keys(%{$files{$file}})) {
    my $fh = $files{$file}{$suffix};
    defined($fh) && defined(fileno($fh)) and close($fh);
    unlink("$tmp_dir/$file.$suffix");
  }
}}

sub alignHashID($) {
  my ($align) = @_;
#  return 23*$$align[START1] + 41*$$align[START2] + 61*$$align[END1] + 83*$$align[END2];
  return "$$align[SEQ1]:$$align[START1]-$$align[END1]=" .
    "$$align[SEQ2]:$$align[START2]-$$align[END2]";
}


# The chain writer lags the chainer by two chains because the full contents
# of neighboring chains must be known.

sub printChainToTemp {
  my ($FH, $prev_chain, $cur_chain, $next_chain, $FH_DM) = @_;
  return unless defined $cur_chain;

  my $type = $$cur_chain[0][ORIGIN];
  my $fh = (($type == 2) && $FH_DM) ? $FH_DM : $FH;
  my ($first_align, $last_align) = ($$cur_chain[0], $$cur_chain[$#$cur_chain]);
  (($type == 3) && ($$first_align[START1] > $$last_align[START1])) ||
      (($type == 2) && $FH_DM &&
      ($$first_align[START2] > $$last_align[START2])) and
    ($first_align, $last_align) = ($last_align, $first_align),
    $type = -$type;
  print $fh join(" ", $type, scalar(@$cur_chain),
    @$first_align[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE],
    @$last_align[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE]);
  if ((($type == 2) || ($type == -2)) && $FH_DM) {
    foreach my $align (@$cur_chain) {
      print $fh " ",
        join(" ", @$align[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE]);
    }
  } elsif ($print_chains) {
    foreach my $align (@$cur_chain) {
      print $fh " ", join(" ", @$align[START1,END1,START2,END2]);
    }
  }
  print $fh "\n";
}

sub chainBase1Hits {
  my ($FH, $hashesDM, $glocal_out_pos, $sort1_pos) = @_;
  my ($out, $cur_align, $prev_align, $cur_chain, $prev_chain, $pre_prev_chain,
    $chain_start_2M, $chain_start_1M1, $cur_seq, $align_peers, $flip_counter,
    @bad_aligns, %base2peers, $prev_seq, $cur_outf, $sort1, $out_dm);
  my $re = '^\s*\((\d+)\s(\d+)\)=\((\d+)\s(\d+)\)\s([-\.\d]+)\s([-+]+)\s' .
    '\[([-\.\d]+)\]\s*s1:\s*(\S+)\s+s2:\s*(\S+)\s*$';

  $out = open_file("gnu_split -a 10 -C 100m - " .
    "'$tmp_dir/$input_files[0].M1.'", "|") or die $Utils::Error;
  $out_dm = open_file("gnu_split -a 10 -C 100m - " .
    "'$tmp_dir/$input_files[0].rawDM.'", "|") or die $Utils::Error;
  while (<$FH>) {
    $cur_align = [];
    (@$cur_align[START1,END1,START2,END2,SCORE,ORIENT,TOTSC,SEQ1,SEQ2]) =
      ($_ =~ /$re/o);
    $$cur_align[START1] && $$cur_align[END1] && $$cur_align[START2] &&
        $$cur_align[END2] && $$cur_align[SCORE] && $$cur_align[ORIENT] or
      (push(@bad_aligns, $_), next);
    $$cur_align[START1] eq $$cur_align[END1] and next; # skip null alignments
    checkAlignCoords($cur_align);
    if (!defined($prev_seq) || ($$cur_align[SEQ1] ne $prev_seq)) {
      my ($pos);
      ($cur_outf, $pos, $sort1) =
        get_index("$input_files[0].sort1", $cur_outf);
      $$sort1_pos{$$cur_align[SEQ1]} = [$cur_outf, $pos];
      $prev_seq = $$cur_align[SEQ1];
    }
# save hits sorted by seq1 coordinates to disk (will be used to expand M2 regs)
    print $sort1 join(" ", @$cur_align[SEQ1,START1,END1]), "\n";
    if ($proflip && defined($flipped_aligns{alignHashID($cur_align)})) {
      my $seq2center = $$sizes2{(keys(%$sizes2))[0]} / 2;
      my $j = $$cur_align[START2];
      $$cur_align[START2] = (2 * $seq2center) - $$cur_align[END2];
      $$cur_align[END2] = (2 * $seq2center) - $j;
      $$cur_align[ORIENT] = ($$cur_align[ORIENT] eq "+") ? "-" : "+";
      $$cur_align[FLIPPED]=1;
      $flip_counter++;
    }

    $$cur_align[HASHID] = alignHashID($cur_align);

    if (!defined($cur_seq) || ($$cur_align[SEQ1] ne $cur_seq)) {
      printChainToTemp($out, $pre_prev_chain, $prev_chain, $cur_chain, $out_dm);
      printChainToTemp($out, $prev_chain, $cur_chain, undef, $out_dm);

      $chain_start_2M = $chain_start_1M1 = $prev_align =
        $pre_prev_chain = $prev_chain = $cur_chain = undef;
      $cur_seq = $$cur_align[SEQ1];
      %base2peers = %{loadBase2Hashes($cur_seq,
        "$input_files[0].gen2.sorted-gen1.glocal-out",
        @{$$glocal_out_pos{$cur_seq}})};
    }

    $align_peers = $base2peers{$$cur_align[HASHID]};
    $$cur_align[ORIGIN] = defined($align_peers) ? 2 : 1;

    if ($chain_start_2M and defined $align_peers and defined $prev_align # continue open 2M chain
        and (($$cur_align[ORIENT] eq "+" and
        $$cur_align[START2] > $$prev_align[END2] and
        $$prev_align[HASHID] eq $$align_peers[0]) or
        ($$cur_align[ORIENT] eq "-" and
        $$cur_align[END2] < $$prev_align[START2] and
        $$prev_align[HASHID] eq $$align_peers[1]) or
        ($$cur_align[FLIPPED] and ($$cur_align[ORIENT] eq "+" and
        $$cur_align[START2] < $$prev_align[END2] and
        $$prev_align[HASHID] eq $$align_peers[0]) or
        ($$cur_align[ORIENT] eq "-" and
        $$cur_align[END2] > $$prev_align[START2] and
        $$prev_align[HASHID] eq $$align_peers[1]))) and
        $$cur_align[ORIENT] eq $$prev_align[ORIENT] and
        $$cur_align[FLIPPED] eq $$prev_align[FLIPPED] and
        $$cur_align[SEQ2] eq $$prev_align[SEQ2] and
        ($$cur_align[START1] > $$prev_align[END1] or
        ($$cur_align[FLIPPED] and
        $$cur_align[START1] > $$prev_align[END1])) and
        abs($$cur_align[END1] - $$chain_start_2M[START1]) < $max_chainlen and
        abs(($$cur_align[ORIENT] eq "+") ?
	  ($$cur_align[END2] - $$chain_start_2M[START2]) :
	  ($$chain_start_2M[END2] - $$cur_align[START2])) < $max_chainlen and
        abs($$cur_align[END1] - $$chain_start_2M[START1]) /
        abs(($$cur_align[ORIENT] eq "+") ?
	  ($$cur_align[END2] - $$chain_start_2M[START2]) :
	  ($$chain_start_2M[END2] - $$cur_align[START2])) < $max_asym and
        abs(($$cur_align[ORIENT] eq "+") ?
	  ($$cur_align[END2] - $$chain_start_2M[START2]) :
	  ($$chain_start_2M[END2] - $$cur_align[START2])) /
        abs($$cur_align[END1] - $$chain_start_2M[START1]) < $max_asym) {
      push(@$cur_chain, $cur_align);
      print $hashesDM join("\t", @$cur_align[SEQ1,SEQ2,HASHID]), "\n";
    } elsif (defined $align_peers) { # start new 2M chain
      printChainToTemp($out, $pre_prev_chain, $prev_chain, $cur_chain, $out_dm);
      $chain_start_2M = $cur_align;
      undef $chain_start_1M1;
      $pre_prev_chain = $prev_chain;
      $prev_chain = $cur_chain;
      $cur_chain = [$cur_align];
      print $hashesDM join("\t", @$cur_align[SEQ1,SEQ2,HASHID]), "\n";
    } elsif ($chain_start_1M1 and defined $prev_align # continue open 1M1 chain
        and ((($$cur_align[ORIENT] eq "+" and
	$$cur_align[START2] > $$prev_align[END2]) or
	($$cur_align[ORIENT] eq "-" and
	$$cur_align[END2] < $$prev_align[START2])) or
	($$cur_align[FLIPPED] and (($$cur_align[ORIENT] eq "+" and
	$$cur_align[START2] < $$prev_align[END2]) or
	($$cur_align[ORIENT] eq "-" and
	$$cur_align[END2] > $$prev_align[START2])))) and
	$$cur_align[ORIENT] eq $$prev_align[ORIENT] and
	$$cur_align[FLIPPED] eq $$prev_align[FLIPPED] and
	$$cur_align[SEQ2] eq $$prev_align[SEQ2] and
	($$cur_align[START1] > $$prev_align[END1] or
	($$cur_align[FLIPPED] and
	$$cur_align[START1] > $$prev_align[END1])) and
	abs($$cur_align[END1] - $$chain_start_1M1[START1]) < $max_chainlen and
	abs(($$cur_align[ORIENT] eq "+") ?
	  ($$cur_align[END2] - $$chain_start_1M1[START2]) :
	  ($$chain_start_1M1[END2] - $$cur_align[START2])) < $max_chainlen and
	abs($$cur_align[END1] - $$chain_start_1M1[START1]) /
	abs(($$cur_align[ORIENT] eq "+") ?
	  ($$cur_align[END2] - $$chain_start_1M1[START2]) :
	  ($$chain_start_1M1[END2] - $$cur_align[START2])) < $max_asym and
	abs(($$cur_align[ORIENT] eq "+") ?
	  ($$cur_align[END2] - $$chain_start_1M1[START2]) :
	  ($$chain_start_1M1[END2] - $$cur_align[START2])) /
	abs($$cur_align[END1] - $$chain_start_1M1[START1]) < $max_asym) {
      push(@$cur_chain, $cur_align);
    } else { # start new 1M1 chain
      printChainToTemp($out, $pre_prev_chain, $prev_chain, $cur_chain, $out_dm);
      $chain_start_1M1 = $cur_align;
      undef $chain_start_2M;
      $pre_prev_chain = $prev_chain;
      $prev_chain = $cur_chain;
      $cur_chain = [$cur_align];
    }
    $prev_align = $cur_align;
  }
  defined($sort1) && defined(fileno($sort1)) and close($sort1);
  printChainToTemp($out, $pre_prev_chain, $prev_chain, $cur_chain, $out_dm);
  printChainToTemp($out, $prev_chain, $cur_chain, undef, $out_dm);
  close($out);
  close($out_dm);
  print "$0: Single-sequence flip mode: " . ($flip_counter + 0) .
    " gen1base hits backflipped\n" if $debug and $proflip;
  warn "$0: Warning: " . scalar(@bad_aligns) .
    " bad SLAGAN alignments discarded" if @bad_aligns > 0;
}


# Input is base 2, i.e. (start2 end2)=(start1 end1)...

sub chainBase2Hits {
  my ($FH, $hashes_pos, $sort2_pos, $FH_DM) = @_;
  my ($out, $cur_align, $prev_align, $cur_chain, $prev_chain, $pre_prev_chain,
    $chain_start_2M, $chain_start_1M2, $cur_seq, $align_is_2M, $flip_counter,
    @bad_aligns, %aligns2M, $prev_seq, $cur_outf, $sort2, $out_dm);
  my $re = '^\s*\((\d+)\s(\d+)\)=\((\d+)\s(\d+)\)\s([-\.\d]+)\s([-+]+)\s' .
    '\[([-\.\d]+)\]\s*s1:\s*(\S+)\s+s2:\s*(\S+)\s*$';

  $out = open_file("gnu_split -a 10 -C 100m - " .
    "'$tmp_dir/$input_files[0].M2.'", "|") or die $Utils::Error;
  $out_dm = open_file("gnu_split -a 10 -C 100m - " .
    "'$tmp_dir/$input_files[0].DM.'", "|") or die $Utils::Error;
  while (<$FH>) {
    $cur_align = [];
    (@$cur_align[START2,END2,START1,END1,SCORE,ORIENT,TOTSC,SEQ2,SEQ1]) =
      ($_ =~ /$re/o);
    $$cur_align[START2] && $$cur_align[END2] && $$cur_align[START1] &&
        $$cur_align[END1] && $$cur_align[SCORE] && $$cur_align[ORIENT] or
      (push(@bad_aligns, $_), next);
    $$cur_align[START2] == $$cur_align[END2] and next; # skip null alignments
    checkAlignCoords($cur_align);
    if (!defined($prev_seq) || ($$cur_align[SEQ2] ne $prev_seq)) {
      my ($pos);
      ($cur_outf, $pos, $sort2) =
        get_index("$input_files[0].sort2", $cur_outf);
      $$sort2_pos{$$cur_align[SEQ2]} = [$cur_outf, $pos];
      $prev_seq = $$cur_align[SEQ2];
    }
# save hits sorted by seq2 coordinates to disk (will be used to expand M1 regs)
    print $sort2 join(" ", $$cur_align[SEQ2], $$cur_align[START2],
      $$cur_align[END2]), "\n";
    if ($proflip && defined($flipped_aligns{alignHashID($cur_align)})) {
      my $seq2center = $$sizes2{(keys(%$sizes2))[0]} / 2;
      my $j = $$cur_align[START2];
      $$cur_align[START2] = (2 * $seq2center) - $$cur_align[END2];
      $$cur_align[END2] = (2 * $seq2center) - $j;
      $$cur_align[ORIENT] = ($$cur_align[ORIENT] eq "+") ? "-" : "+";
      $$cur_align[FLIPPED] = 1;
      $flip_counter++;
    }

    $$cur_align[HASHID] = alignHashID($cur_align);

    if ($$cur_align[SEQ2] ne $cur_seq) {
      printChainToTemp($out, $pre_prev_chain, $prev_chain, $cur_chain)
        if $$prev_chain[0][ORIGIN] == 3;# and not defined $cur_seq;
      printChainToTemp($out, $prev_chain, $cur_chain, undef)
        if $$cur_chain[0][ORIGIN] == 3;# and not defined $cur_seq;
      $chain_start_1M2 = $prev_align = $pre_prev_chain = $prev_chain =
        $cur_chain = undef;
      $cur_seq = $$cur_align[SEQ2];
      %aligns2M = %{load2MHashes($cur_seq, "$input_files[0].gen2.hashesDM",
        @{$$hashes_pos{$cur_seq}})};
    }
    $$cur_align[ORIGIN] = defined($aligns2M{$$cur_align[HASHID]}) ? 2 : 3;
    split_dm_chains($FH_DM, $out_dm, $cur_align);
    if (defined($aligns2M{$$cur_align[HASHID]})) { # align is 2M
      my $prev_ch_last_al = $prev_chain ?
        $$prev_chain[scalar(@$prev_chain)-1] : [];
      printChainToTemp($out, $pre_prev_chain, $prev_chain, $cur_chain)
        if $$prev_chain[0][ORIGIN] == 3;
      undef $chain_start_1M2; # close 1M2 chain
      $chain_start_2M = $cur_align;
      $pre_prev_chain = $prev_chain;
      $prev_chain = $cur_chain;
      $cur_chain = [$cur_align];
    } elsif ($chain_start_1M2 # continue open 1M2 chain
        and ((($$cur_align[ORIENT] eq "+" and
        $$cur_align[START1] > $$prev_align[END1]) or
	($$cur_align[ORIENT] eq "-" and
	$$cur_align[END1] < $$prev_align[START1])) or
	($$cur_align[FLIPPED] and (($$cur_align[ORIENT] eq "+" and
	$$cur_align[START1] < $$prev_align[END1]) or
	($$cur_align[ORIENT] eq "-" and
	$$cur_align[END1] > $$prev_align[START1])))) and
	$$cur_align[ORIENT] eq $$prev_align[ORIENT] and
	$$cur_align[SEQ1] eq $$prev_align[SEQ1] and
	$$cur_align[FLIPPED] == $$prev_align[FLIPPED] and
	($$cur_align[START2] > $$prev_align[END2] or
	($$cur_align[FLIPPED] and
	$$cur_align[START2] < $$prev_align[END2])) and
	abs(($$cur_align[ORIENT] eq "+") ?
	  ($$cur_align[END1] - $$chain_start_1M2[START1]) :
	  ($$chain_start_1M2[END1] - $$cur_align[START1])) < $max_chainlen and
	abs($$cur_align[END2] - $$chain_start_1M2[START2]) < $max_chainlen and
	abs(($$cur_align[ORIENT] eq "+") ?
	  ($$cur_align[END1] - $$chain_start_1M2[START1]) :
	  ($$chain_start_1M2[END1] - $$cur_align[START1])) /
	abs($$cur_align[END2] - $$chain_start_1M2[START2]) < $max_asym and
	abs($$cur_align[END2] - $$chain_start_1M2[START2]) /
	abs(($$cur_align[ORIENT] eq "+") ?
	  ($$cur_align[END1] - $$chain_start_1M2[START1]) :
	  ($$chain_start_1M2[END1] - $$cur_align[START1])) < $max_asym) {
      push(@$cur_chain, $cur_align);
    } else { # start new 1M2 chain
      my $prev_ch_last_al = $prev_chain ?
        $$prev_chain[scalar(@$prev_chain)-1] : [];
      printChainToTemp($out, $pre_prev_chain, $prev_chain, $cur_chain)
        if $$prev_chain[0][ORIGIN] == 3;
      $chain_start_1M2 = $cur_align;
      $pre_prev_chain = $prev_chain;
      $prev_chain = $cur_chain;
      $cur_chain = [$cur_align];
    }
    $prev_align = $cur_align;
  }
  defined($sort2) && defined(fileno($sort2)) and close($sort2);
  my $prev_ch_last_al = $prev_chain ?
    $$prev_chain[scalar(@$prev_chain)-1] : [];
  printChainToTemp($out, $pre_prev_chain, $prev_chain, $cur_chain)
    if $$prev_chain[0][ORIGIN] == 3;
  printChainToTemp($out, $prev_chain, $cur_chain, undef)
    if $$cur_chain[0][ORIGIN] == 3;
  close($out);
  flush_dm_chains($out_dm);
  close($out_dm);
  print "$0: Single-sequence flip mode: " . ($flip_counter + 0) .
    " gen2base hits backflipped\n" if $debug and $proflip;
  warn "$0: Warning: " . scalar(@bad_aligns) .
    " bad SLAGAN alignments discarded" if @bad_aligns > 0;
}

{
my (@chains, @next_chain);

sub split_dm_chains {
  my ($in, $out, $align) = @_;

  get_chains($in, $out, $align);
  split_chains($out, $align);
}

sub get_chains {
  my ($in, $out, $align) = @_;
  my ($line);

  LOOP1: while (1) {
    while (!@next_chain) {
      defined($line = <$in>) or last LOOP1;
      @next_chain = split(" ", $line);
    }
    my $rc = $next_chain[7] cmp $$align[SEQ2];
    ($rc > 0) || (($rc == 0) && ($next_chain[4] > $$align[END2])) and last;
    ($rc < 0) || ($next_chain[13] < $$align[START2]) and
      save_chain($out, \@next_chain), @next_chain = (), next;
    my (@first_align, @last_align);
    (@first_align[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE],
      @last_align[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE]) =
      @next_chain[2..17];
    my @chain = ([@next_chain], \@first_align, \@last_align, {});
    for (my $i = 18; $i < @next_chain; $i += 8) {
      my (@cur_align);
      @cur_align[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE] =
        @next_chain[$i..($i + 7)];
      $chain[3]{alignHashID(\@cur_align)} = 1;
    }
    push(@chains, \@chain);
    @next_chain = ();
  }
}

sub split_chains {
  my ($out, $align) = @_;

  for (my $i = 0; $i < @chains; $i++) {
    ($chains[$i][1][SEQ2] ne $$align[SEQ2]) ||
        ($chains[$i][2][END2] < $$align[START2]) and
      save_chain($out, $chains[$i][0]),
      splice(@chains, $i, 1),
      $i--,
      next;
    ($chains[$i][1][START2] > $$align[END2]) ||
      (($$align[ORIGIN] == 2) && $chains[$i][3]{$$align[HASHID]}) and next;
    my ($chain1, $chain2) = split_chain($chains[$i], $align);
    $chain1 and save_chain($out, $$chain1[0]);
    splice(@chains, $i, 1, $chain2 ? ($chain2) : ());
    $chain2 or $i--;
  }
}

sub split_chain {
  my ($chain, $align) = @_;
  my @chains = ([[$$chain[0][0], 0], [], [], {}],
    [[$$chain[0][0], 0], [], [], {}]);

  for (my $i = 18; $i < @{$$chain[0]}; $i += 8) {
    my (@cur_align);
    @cur_align[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE] =
      @{$$chain[0]}[$i..($i + 7)];
    my $n = ($cur_align[END2] < $$align[START2]) ? 0 : 1;
    push(@{$chains[$n][0]},
      @cur_align[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE]);
    $chains[$n][0][1]++;
    $n and $chains[$n][3]{alignHashID(\@cur_align)} = 1;
  }
  for (my $n = 0; $n < 2; $n++) {
    $chains[$n][0][1] or next;
    @{$chains[$n][1]}[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE] =
      @{$chains[$n][0]}[2..9];
    @{$chains[$n][2]}[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE] =
      @{$chains[$n][0]}[($#{$chains[$n][0]} - 7)..$#{$chains[$n][0]}];
    $chains[$n][0][0] < 0 and
      ($chains[$n][1], $chains[$n][2]) = ($chains[$n][2], $chains[$n][1]);
    splice(@{$chains[$n][0]}, 2, 0,
      @{$chains[$n][1]}[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE],
      @{$chains[$n][2]}[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE]);
  }
  return ($chains[0][0][1] ? $chains[0] : undef,
    $chains[1][0][1] ? $chains[1] : undef);
}

sub flush_dm_chains {
  my ($out) = @_;

  for (my $i = 0; $i < @chains; $i++) { save_chain($out, $chains[$i][0]); }
  @chains = ();
}

sub save_chain {
  my ($out, $chain) = @_;
  my ($first_align, $last_align) = ([], []);

  (@$first_align[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE],
    @$last_align[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE]) =
    @$chain[2..17];
  $$chain[0] < 0 and
    ($first_align, $last_align) = ($last_align, $first_align),
    $$chain[0] = -$$chain[0];
  print $out join(" ", @$chain[0,1],
    @$first_align[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE],
    @$last_align[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE]);
  if ($print_chains) {
    for (my $i = 18; $i < @$chain; $i += 8) {
      my (@cur_align);
      @cur_align[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE] =
        @$chain[$i..($i + 7)];
      print $out " ", join(" ", @cur_align[START1,END1,START2,END2]);
    }
  }
  print $out "\n";
}}

# Input: file with lines of the form "seq1 seq2 hash" (seq2 should be
# the same per file)
# Output: hash(key->align hash ID, value->1). Input file is deleted.

sub load2MHashes {
  my ($seq_name, $file, $suffix, $pos) = @_;
  my ($in, %hashes);

  $in = open_db_file($file, $suffix, $pos) or return {};
  while (<$in>) {
    /^\S+\s+(\S+)\s+(\S+)$/o;
    $1 eq $seq_name or last;
    defined($hashes{$2}) and
      warn("Hash collision in \"$_\" vs. \"".$hashes{$2}."\"");
    $hashes{$2} = 1;
  }
  return \%hashes;
}

# Input: file with gen2base alignments which should have the same seq1
# ordered by start2 or not exist
# Output: hash(key->align hash ID,
# value->[prev align hash ID, next align hash ID]). Input file is deleted.
# Input is base 2, i.e. (start2 end2)=(start1 end1)...

sub loadBase2Hashes {
  my ($seq_name, $file, $suffix, $pos) = @_;
  my ($prev_align, $cur_align, $next_align, %hashes, $in);
  my $re = '^\s*\((\d+)\s(\d+)\)=\((\d+)\s(\d+)\)\s.*' .
    's1:\s*(\S+)\s+s2:\s*(\S+)\s*$';
  
  $in = open_db_file($file, $suffix, $pos) or return {};
  # Scan 1 line ahead because the next align must also be seen
  while (<$in>) {
    $next_align = [];
    # Hits are gen2base
    (@$next_align[START2,END2,START1,END1,SEQ2,SEQ1]) = ($_ =~ /$re/o);
    $$next_align[SEQ1] eq $seq_name or last;
    checkAlignCoords($next_align);
    $$next_align[HASHID] = alignHashID($next_align);
    if ($cur_align) {
      exists($hashes{$$cur_align[HASHID]}) and
        warn("LB2H: Hash collision in \"$_\"");
      $hashes{$$cur_align[HASHID]} =
        [$prev_align ? $$prev_align[HASHID] : 1,
        $next_align ? $$next_align[HASHID] : 1];
    }
    $prev_align = $cur_align;
    $cur_align = $next_align;
  }
  if ($cur_align) {
    exists($hashes{$$cur_align[HASHID]}) and
      warn("LB2H: Hash collision in \"$_\"");
    $hashes{$$cur_align[HASHID]} =
      [$prev_align ? $$prev_align[HASHID] : 1, undef];
  }
  return \%hashes;
}


# Load chained regions and expand them according to the expansion rules,
# then print them out and display some chain statistics

sub postProcessRegions {
  my ($sort1_pos, $sort2_pos) = @_;
  local (*OUT, *RH1, *WH1, *RH2, *WH2, *RH3, *WH3);
  my ($first_align, $last_align, $type, $num_aligns, $sort_pid1, $sort_pid2,
    $sort_pid3, @line, @min_lengths, @max_lengths, @means, @pos_counts,
    @neg_counts);

  # sort on seq1, start1
  $sort_pid1 = open2(\*RH1, \*WH1, "sort -k 7,7 -k 3,3n");
  # sort on seq2, start2
  $sort_pid2 = open2(\*RH2, \*WH2, "sort -k 8,8 -k 5,5n");
  open(OUT, "> $outfile") or
    die("$0: Could not open output file $outfile for writing: $!");
  foreach my $file (safe_glob("^$input_files[0]" . '\.DM\.\w{10}$',
      $tmp_dir), safe_glob("^$input_files[0]" . '\.M1\.\w{10}$', $tmp_dir),
      safe_glob("^$input_files[0]" . '\.M2\.\w{10}$', $tmp_dir)) {
    copy("$tmp_dir/$file", \*WH1);
    $nodelete or unlink("$tmp_dir/$file");
  }
  close(WH1);
  expandSeq1(\*RH1, \*WH2, $sort1_pos);
  close(RH1);
  waitpid($sort_pid1, 0);
  close(WH2);
  $nodelete or rm_db_files("$input_files[0].sort1");
  expandSeq2(\*RH2, \*OUT, $sort2_pos);
  close(RH2);
  waitpid($sort_pid2, 0);
  close(OUT);
  $nodelete or rm_db_files("$input_files[0].sort2");
}

{
my (%cur_seq, %fh);

sub expand_chain {
  my ($seq, $start, $end, $file, $pos, $second_seq) = @_;
  my ($fh, $cur_pos, $prev_pos, $prev_hit_e, $exp_start, $exp_end, $expand);

  if (!defined($cur_seq{$file}) || ($cur_seq{$file} ne $seq)) {
    $fh{$file} = open_db_file($file, @{$$pos{$seq}});
    $cur_seq{$file} = $seq;
  }
  $fh = $fh{$file};
  if (defined($fh)) {
    $prev_pos = $cur_pos = tell($fh);
    while (<$fh>) {
      my ($hit_seq, $hit_s, $hit_e) = split();
      $hit_seq ne $seq and last;
      if (!defined($exp_start)) {
        if ($hit_e >= $start) {
          $exp_start = defined($prev_hit_e) ? ($prev_hit_e + 1) : 1;
          seek($fh, $cur_pos, 0);
        } else {
          $prev_pos = $cur_pos;
          $prev_hit_e = $hit_e;
          $cur_pos = tell($fh);
        }
      } elsif ($hit_s > $end) {
        $exp_end = $hit_s - 1;
        last;
      }
    }
    seek($fh, $prev_pos, 0);
  }
  defined($exp_start) or
    $exp_start = defined($prev_hit_e) ? ($prev_hit_e + 1) : 1;
  defined($exp_end) or
    $exp_end = $second_seq ? $$sizes2{$seq} : $$sizes1{$seq};
  $expand = $start - $exp_start;
  $expand > $max_expand_len and $expand = $max_expand_len;
  $exp_start = $start - $expand;
  $expand = $exp_end - $end;
  $expand > $max_expand_len and $expand = $max_expand_len;
  $exp_end = $end + $expand;
  return ($exp_start, $exp_end);
}}

# Input: chains ordered by seq1, start1
# Output: chains expanded on seq1

sub expandSeq1($$$) {
  my ($RH, $WH, $sort1_pos) = @_;
  my ($first_align, $last_align, $type, $num_aligns,
    $cur_seq, $preexpand1, $postexpand1,
    $prev_chain, $cur_chain, $next_chain);

  while (<$RH>) {
    chomp();
    my @line = split();
    @line or next;

    $first_align = [];
    $last_align = [];
    ($type, $num_aligns,
      @$first_align[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE],
      @$last_align[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE]) = @line;
    splice(@line, 0, 18);
    $type < 0 and 
      ($first_align, $last_align) = ($last_align, $first_align),
      $type = -$type;
    $$first_align[CHALO1] = ($$first_align[START1] < $$last_align[START1]) ?
      $$first_align[START1] : $$last_align[START1];
    $$first_align[CHAHI1] = ($$first_align[END1] > $$last_align[END1]) ?
      $$first_align[END1] : $$last_align[END1];

    # special expansion for M2 regions
    if ($type == 3) {
      ($$first_align[CHALO1E], $$first_align[CHAHI1E]) =
        expand_chain($$first_align[SEQ1], $$first_align[CHALO1],
        $$first_align[CHAHI1], "$input_files[0].sort1", $sort1_pos, 0);
      print $WH join(" ", $type, $num_aligns,
        @$first_align[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE],
        @$last_align[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE],
        @$first_align[CHALO1,CHAHI1,CHALO1E,CHAHI1E], @line), "\n";
      next;
    }

    $prev_chain = $cur_chain;
    $cur_chain = $next_chain;

#    $$first_align[CHALO1] = ($$first_align[START1] < $$last_align[START1] ?
#      $$first_align[START1] : $$last_align[START1]);
#    $$first_align[CHAHI1] = ($$first_align[END1] > $$last_align[END1] ?
#      $$first_align[END1] : $$last_align[END1]);

    $$first_align[START1] > $$last_align[START1] and die;
    $$first_align[END1] > $$last_align[END1] and die;

    $next_chain = [$first_align, $last_align, $type, $num_aligns, \@line];

    expSeq1Reg($WH, $prev_chain, $cur_chain, $next_chain);
  }
  expSeq1Reg($WH, $cur_chain, $next_chain, undef);
}


sub expSeq1Reg {
  my ($WH, $prev_chain, $cur_chain, $next_chain) = @_;
  my ($preexpand1, $postexpand1, $type, $first_align, $last_align);

  defined($cur_chain) or return;
  $preexpand1 = $$cur_chain[0][CHALO1] - 1 -
    ((defined($prev_chain) &&
    ($$prev_chain[0][SEQ1] eq $$cur_chain[0][SEQ1])) ?
    $$prev_chain[0][CHAHI1] : 0);
  ($preexpand1 > $max_expand_len) || ($preexpand1 < 0) and
    $preexpand1 = $max_expand_len;
  $postexpand1 = ((defined($next_chain) &&
    ($$next_chain[0][SEQ1] eq $$cur_chain[0][SEQ1])) ?
    ($$next_chain[0][CHALO1] - 1) : $$sizes1{$$cur_chain[0][SEQ1]}) -
    $$cur_chain[0][CHAHI1];
  ($postexpand1 > $max_expand_len) || ($postexpand1 < 0) and
    $postexpand1 = $max_expand_len;
  $$cur_chain[0][CHALO1E] = $$cur_chain[0][CHALO1] - $preexpand1;
  $$cur_chain[0][CHALO1E] < 1 and $$cur_chain[0][CHALO1E] = 1;
  $$cur_chain[0][CHAHI1E] = $$cur_chain[0][CHAHI1] + $postexpand1;
  $$cur_chain[0][CHAHI1E] > $$sizes1{$$cur_chain[0][SEQ1]} and
    $$cur_chain[0][CHAHI1E] = $$sizes1{$$cur_chain[0][SEQ1]};
  $type = $$cur_chain[2];
  ($first_align, $last_align) = ($$cur_chain[0], $$cur_chain[1]);
  $$first_align[START2] > $$last_align[START2] and
    ($first_align, $last_align) = ($last_align, $first_align),
    $type = -$type;
  print $WH join(" ", $type, $$cur_chain[3],
    @$first_align[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE],
    @$last_align[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE],
    @{$$cur_chain[0]}[CHALO1,CHAHI1,CHALO1E,CHAHI1E], @{$$cur_chain[4]}), "\n";
}

# Input: chains ordered by seq2, start2
# Output: chains expanded on seq1 and seq2 (final output)

sub expandSeq2($$$) {
  my ($RH, $WH, $sort2_pos) = @_;
  my ($first_align, $last_align, $type, $num_aligns,
    $cur_seq, $preexpand1, $postexpand1, $preexpand2, $postexpand2,
    $prev_chain, $cur_chain, $next_chain);

  while (<$RH>) {
    chomp();
    my @line = split();
    @line or next;

    $first_align = [];
    $last_align = [];
    ($type, $num_aligns,
      @$first_align[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE],
      @$last_align[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE],
      @$first_align[CHALO1,CHAHI1,CHALO1E,CHAHI1E]) = @line;
    splice(@line, 0, 22);
    $type < 0 and 
      ($first_align, $last_align) = ($last_align, $first_align),
      ($$first_align[CHALO1], $$first_align[CHAHI1], $$first_align[CHALO1E],
        $$first_align[CHAHI1E]) = ($$last_align[CHALO1], $$last_align[CHAHI1],
        $$last_align[CHALO1E], $$last_align[CHAHI1E]),
      $type = -$type;
    $$first_align[CHALO2] = ($$first_align[START2] < $$last_align[START2]) ?
      $$first_align[START2] : $$last_align[START2];
    $$first_align[CHAHI2] = ($$first_align[END2] > $$last_align[END2]) ?
      $$first_align[END2] : $$last_align[END2];

    # special expansion for M1 regions
    if ($type == 1) {
      ($$first_align[CHALO2E], $$first_align[CHAHI2E]) =
        expand_chain($$first_align[SEQ2], $$first_align[CHALO2],
        $$first_align[CHAHI2], "$input_files[0].sort2", $sort2_pos, 1);
      my @lines = (join(" ", $type, $num_aligns,
        @$first_align[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE],
        @$last_align[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE],
        @$first_align[CHALO1,CHAHI1,CHALO1E,CHAHI1E,CHALO2,CHAHI2,
        CHALO2E,CHAHI2E], @line) . "\n");
      finalExpand($WH, \@lines);
      next;
    }

    $prev_chain = $cur_chain;
    $cur_chain = $next_chain;

    $next_chain = [$first_align, $last_align, $type, $num_aligns, \@line];

    expSeq2Reg($WH, $prev_chain, $cur_chain, $next_chain);
  }
  expSeq2Reg($WH, $cur_chain, $next_chain, undef);
}

sub expSeq2Reg {
  my ($WH, $prev_chain, $cur_chain, $next_chain) = @_;
  my ($preexpand1, $postexpand1, $preexpand2, $postexpand2, $line, @lines);

  defined($cur_chain) or return;
  $preexpand1 = $$cur_chain[0][CHALO1] - $$cur_chain[0][CHALO1E];
  $postexpand1 = $$cur_chain[0][CHAHI1E] - $$cur_chain[0][CHAHI1];
  $preexpand2 = $$cur_chain[0][CHALO2] - 1 -
    ((defined($prev_chain) &&
    ($$prev_chain[0][SEQ2] eq $$cur_chain[0][SEQ2])) ?
    $$prev_chain[0][CHAHI2] : 0);
  ($preexpand2 > ($preexpand1 * $expand_factor)) and
    $preexpand2 = $preexpand1 * $expand_factor;
  ($preexpand2 > $max_expand_len) || ($preexpand2 < 0) and
    $preexpand2 = $max_expand_len;
  ($preexpand1 > $preexpand2 * $expand_factor) and
    $preexpand1 = $preexpand2 * $expand_factor;
  $preexpand1 > $max_expand_len and $preexpand1 = $max_expand_len;
  $postexpand2 = ((defined($next_chain) &&
    ($$next_chain[0][SEQ2] eq $$cur_chain[0][SEQ2])) ?
    ($$next_chain[0][CHALO2] - 1) : $$sizes2{$$cur_chain[0][SEQ2]}) -
    $$cur_chain[0][CHAHI2];
  ($postexpand2 > $postexpand1 * $expand_factor) and
    $postexpand2 = $postexpand1 * $expand_factor;
  ($postexpand2 > $max_expand_len) || ($postexpand2 < 0) and
    $postexpand2 = $max_expand_len;
  ($postexpand1 > $postexpand2 * $expand_factor) and
    $postexpand1 = $postexpand2 * $expand_factor;
  $postexpand1 > $max_expand_len and $postexpand1 = $max_expand_len;
  $$cur_chain[0][CHALO1E] = $$cur_chain[0][CHALO1] - $preexpand1;
  $$cur_chain[0][CHALO1E] < 1 and $$cur_chain[0][CHALO1E] = 1;
  $$cur_chain[0][CHAHI1E] = $$cur_chain[0][CHAHI1] + $postexpand1;
  $$cur_chain[0][CHAHI1E] > $$sizes1{$$cur_chain[0][SEQ1]} and
    $$cur_chain[0][CHAHI1E] = $$sizes1{$$cur_chain[0][SEQ1]};
  $$cur_chain[0][CHALO2E] = $$cur_chain[0][CHALO2] - $preexpand2;
  $$cur_chain[0][CHALO2E] < 1 and $$cur_chain[0][CHALO2E] = 1;
  $$cur_chain[0][CHAHI2E] = $$cur_chain[0][CHAHI2] + $postexpand2;
  $$cur_chain[0][CHAHI2E] > $$sizes2{$$cur_chain[0][SEQ2]} and
    $$cur_chain[0][CHAHI2E] = $$sizes2{$$cur_chain[0][SEQ2]};
  $line = join(" ", $$cur_chain[2], $$cur_chain[3],
    @{$$cur_chain[0]}[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE],
    @{$$cur_chain[1]}[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE],
    @{$$cur_chain[0]}[CHALO1,CHAHI1,CHALO1E,CHAHI1E,CHALO2,CHAHI2,
    CHALO2E,CHAHI2E], @{$$cur_chain[4]}) . "\n";
  @lines = ($line);
  finalExpand($WH, \@lines);
}


sub finalExpReg {
  my ($WH, $prev_chain, $cur_chain, $next_chain) = @_;
  my ($preexpand1, $postexpand1, $preexpand2, $postexpand2);

  defined($cur_chain) or return;
  if ($$cur_chain[2] == 1) { # M1: expand in seq1 on seq2 expands * factor only
    $preexpand1 = $$cur_chain[0][CHALO1] - $$cur_chain[0][CHALO1E];
    $preexpand2 = $$cur_chain[0][CHALO2] - $$cur_chain[0][CHALO2E];
    ($preexpand2 > ($preexpand1 * $expand_factor)) and
      $preexpand2 = $preexpand1 * $expand_factor;
    $preexpand2 > $max_expand_len and $preexpand2 = $max_expand_len;
    ($preexpand1 > $preexpand2 * $expand_factor) and
      $preexpand1 = $preexpand2 * $expand_factor;
    $preexpand1 > $max_expand_len and $preexpand1 = $max_expand_len;
    $postexpand1 = $$cur_chain[0][CHAHI1E] - $$cur_chain[0][CHAHI1];
    $postexpand2 = $$cur_chain[0][CHAHI2E] - $$cur_chain[0][CHAHI2];
    ($postexpand2 > $postexpand1 * $expand_factor) and
      $postexpand2 = $postexpand1 * $expand_factor;
    $postexpand2 > $max_expand_len and $postexpand2 = $max_expand_len;
    ($postexpand1 > $postexpand2 * $expand_factor) and
      $postexpand1 = $postexpand2 * $expand_factor;
    $postexpand1 > $max_expand_len and $postexpand1 = $max_expand_len;
    $$cur_chain[0][CHALO1E] = $$cur_chain[0][CHALO1] - $preexpand1;
    $$cur_chain[0][CHALO1E] < 1 and $$cur_chain[0][CHALO1E] = 1;
    $$cur_chain[0][CHAHI1E] = $$cur_chain[0][CHAHI1] + $postexpand1;
    $$cur_chain[0][CHAHI1E] > $$sizes1{$$cur_chain[0][SEQ1]} and
      $$cur_chain[0][CHAHI1E] = $$sizes1{$$cur_chain[0][SEQ1]};
    $$cur_chain[0][CHALO2E] = $$cur_chain[0][CHALO2] - $preexpand2;
    $$cur_chain[0][CHALO2E] < 1 and $$cur_chain[0][CHALO2E] = 1;
    $$cur_chain[0][CHAHI2E] = $$cur_chain[0][CHAHI2] + $postexpand2;
    $$cur_chain[0][CHAHI2E] > $$sizes2{$$cur_chain[0][SEQ2]} and
      $$cur_chain[0][CHAHI2E] = $$sizes2{$$cur_chain[0][SEQ2]};
  }
  $$cur_chain[3] or return;
  print $WH join(" ", @{$$cur_chain[0]}[SEQ1,CHALO1E,CHAHI1E], " ",
    @{$$cur_chain[0]}[SEQ2,CHALO2E,CHAHI2E,ORIENT]);
  $no_aligntotals or
    print $WH " (" . ($$cur_chain[2]==1 ?
      "M1, " : $$cur_chain[2]==2 ? "DM, " : "M2, ") . $$cur_chain[3] .
      " aligns)";
  if ($print_chains) {
    my $i = 26;
    while (1) {
      print $WH " [" . $$cur_chain[4][$i] . "-" . $$cur_chain[4][$i + 1] .
        "=" . $$cur_chain[4][$i + 2] . "-" . $$cur_chain[4][$i + 3] . "]";
      @{$$cur_chain[4]} <= ($i + 4) and last;
      $i += 4;
    }
  }
  print $WH "\n";
}


sub finalExpand($$) {
  my ($WH, $lines) = @_;
  my ($first_align, $last_align, $type, $num_aligns,
    $cur_seq, $preexpand1, $postexpand1, $preexpand2, $postexpand2,
    $prev_chain, $cur_chain, $next_chain);
  my %stats;

  foreach (@$lines) {
    chomp();
    my @line = split();
    @line or next;

    $prev_chain = $cur_chain;
    $cur_chain = $next_chain;

    $first_align = [];
    $last_align = [];
    ($type, $num_aligns,
      @$first_align[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE],
      @$last_align[START1,END1,START2,END2,SEQ1,SEQ2,ORIENT,SCORE],
      @$first_align[CHALO1,CHAHI1,CHALO1E,CHAHI1E,CHALO2,CHAHI2,
      CHALO2E,CHAHI2E]) = @line;

    if ($type == 1) {
      $$first_align[CHALO2] = ($$first_align[START2] < $$last_align[START2]) ?
        $$first_align[START2] : $$last_align[START2];
      $$first_align[CHAHI2] = ($$first_align[END2] > $$last_align[END2]) ?
        $$first_align[END2] : $$last_align[END2];
    } elsif ($type == 3) {
      $$first_align[CHALO1] = ($$first_align[START1] < $$last_align[START1]) ?
        $$first_align[START1] : $$last_align[START1];
      $$first_align[CHAHI1] = ($$first_align[END1] > $$last_align[END1]) ?
        $$first_align[END1] : $$last_align[END1];
    }

    $next_chain = [$first_align, $last_align, $type, $num_aligns, \@line];

    finalExpReg($WH, $prev_chain, $cur_chain, $next_chain);

    defined($cur_chain) or next;

    if ($debug || $print_stats) {
      if ($type == 1) {
        $$cur_chain[0][ORIENT] eq "+" ? $stats{"M1+"}++ : $stats{"M1-"}++;
	($stats{"M1min"} > $num_aligns) || !defined($stats{"M1min"}) and
          $stats{"M1min"} = $num_aligns;
        ($stats{"M1max"} < $num_aligns) || !defined($stats{"M1max"}) and
          $stats{"M1max"} = $num_aligns;
        $stats{"M1mean"} += $num_aligns;
      } elsif ($type == 2) {
        $$cur_chain[0][ORIENT] eq "+" ? $stats{"DM+"}++ : $stats{"DM-"}++;
	($stats{"DMmin"} > $num_aligns) || !defined($stats{"DMmin"}) and
          $stats{"DMmin"} = $num_aligns;
        ($stats{"DMmax"} < $num_aligns) || !defined($stats{"DMmax"}) and
          $stats{"DMmax"} = $num_aligns;
        $stats{"DMmean"} += $num_aligns;
      } else {
        $$cur_chain[0][ORIENT] eq "+" ? $stats{"M2+"}++ : $stats{"M2-"}++;
	($stats{"M2min"} > $num_aligns) || !defined($stats{"M2min"}) and
          $stats{"M2min"} = $num_aligns;
	($stats{"M2max"} < $num_aligns) || !defined($stats{"M2max"}) and
          $stats{"M2max"} = $num_aligns;
        $stats{"M2mean"} += $num_aligns;
      }
    }
  }
  finalExpReg($WH, $cur_chain, $next_chain, undef);

  if ($debug || $print_stats) {
    foreach my $i ("DM", "M1", "M2") {
      ($stats{$i."+"} + $stats{$i."-"} == 0) or
        $stats{$i . "mean"} /= ($stats{$i . "+"} + $stats{$i . "-"});
      print "$i: " . ($stats{$i . "+"} + $stats{$i . "-"}) . " chains (" .
        $stats{$i . "+"} . "+, " . $stats{$i . "-"} . "-); " .
        "length min " . $stats{$i . "min"} . ", avg " . $stats{$i . "mean"} .
	", max " . $stats{$i . "max"} . "\n";
    }
  }
}

# Called only in a "$0 worker" invocation

sub workerRun($$$$) {
  my ($tar_file, $score_file, $SLAGAN, $debug) = @_;
  my ($tmp_dir, $io_dir) = ($worker_tmp_dir, getcwd());
  local *FH;

  useIf 1, "Utils";
  mkdir($tmp_dir) or
    die("$0 (worker): Could not create directory $tmp_dir: ".$!);
  $debug and
    print "$0 (worker): Version $VERSION started " . localtime() . "\n";
  $debug and
    print "$0 (worker): Jobfile=$tar_file, scorefile=$score_file, " .
      "tmpdir=$tmp_dir, iodir=$io_dir, SLAGAN=$SLAGAN\n";
  my @files = `tar -tf '$tar_file'`;
  $? and warn("$0 (worker): Error listing contents of $tar_file");
  `tar -xf '$tar_file' -C '$tmp_dir' -m`;
  $? and warn("$0 (worker): Error extracting $tar_file");
  $tar_file =~ s/\.tar$//;
  foreach my $file (@files) {
    chomp $file;
    system("'$SLAGAN' '$tmp_dir/$file' '$score_file' ".
      "> '$tmp_dir/$file.glocal-out' ".
      "2> '$tmp_dir/$file.glocal-err'");
    my ($prefix, $genome, $seq_name) =
      ($file =~ /^(.+)\.gen(\d)base\.(.+)\.chaos$/o);
    if (-s "$tmp_dir/$file.glocal-out") {
      system("cat '$tmp_dir/$file.glocal-out' >> " .
        "'$prefix.gen$genome.$tar_file.results'");
    } else {
      warn "$0 (worker): Empty output from SLAGAN for genome $genome, " .
        "sequence $seq_name";
    }
  }
  rmtree $tmp_dir;
  my @cores = Utils::safe_glob('^core\.');
  if (@cores) {
    warn("$0 (worker): WARNING: $SLAGAN crashed " .
      scalar(@cores) . " times");
    $nodelete or unlink(@cores);
  }
}


# Interrupt handler
sub dequeueClustJobs($) {
  print "\n$0: Received SIG".$_[0].". Cleaning up... ";
  if ($clust_run_pid) {
    # send SIGQUIT to clust_run so it can dequeue cluster jobs
    kill "QUIT", $clust_run_pid;
  }
  unless ($debug or $nodelete) {
    print "Removing job files...";
    foreach my $i (1..$num_jobs) {
      unlink "$submit_dir/JOB$i.tar";
      unlink(map { "$submit_dir/$_" }
        safe_glob("JOB$i" . '\.results$', $submit_dir));
      unlink "$submit_dir/CLUSTER_JOB_MESSAGES.$i";
      unlink "$submit_dir/CLUSTER_JOB_ERRMSG.$i";
    }

#    unlink "$tmp_dir/$input_glob.chaos";
    foreach my $file (@input_files) {
      unlink("$tmp_dir/$file.chaos");
    }
    unlink("$tmp_dir/CLUSTER_JOB_PARAMS", "$submit_dir/score_file");
    if ($ARGV[0] eq "worker") {
      rmtree($tmp_dir);
    } elsif (!$tmp_existed) {
      rmtree($tmp_dir);
    }
  }
  print "\n";
  exit(1);
}


# Retrieve sequence length data from GPDB

sub get_all_seqs($$) {
  my ($dbh, $genome) = @_;
  my ($dset, $annot_db, $family, $check_chroms, %sizes, $chroms, @real_chroms,
  $ctgs);

  ($dset, $annot_db, $family) = ($genome =~ /^\d+$/o) ?
    ($genome + 0, ($dbh->get_data_set($genome))[4,14]) :
    ($dbh->get_family_dset($genome))[0,4,14];
  $debug and
    print "$0: Genome $genome, dataset $dset, annotation db \"$annot_db\", " .
      "family \"$family\"\n";
  $annot_db and $check_chroms = 1;
  if ($check_chroms) {
    $chroms = $dbh->get_chroms(($dbh->get_data_set($dset))[2]);
    foreach my $chrom (@$chroms) {
      $$chrom[1] == 1 or next;
      my $name = "chr$$chrom[2]";
      my ($chr_id, $chr_type, $ctg_id, $size) =
        $dbh->find_seq($name, $dset, $annot_db);
      $chr_id and $sizes{$name} = $size;
    }
  }
  $ctgs = $dbh->selectcol("SELECT name FROM dset$dset\_contigs " .
              "WHERE name is not null and name != ? group by name", undef, "");
  foreach my $ctg (@$ctgs) {
    $sizes{$ctg} = $dbh->get_contig_size($dset, $ctg);
  }
  return \%sizes;
}


sub alnum {
  my ($i);
  my ($len1, $len2) = (length($a), length($b));
  for ($i = 0; ($i < $len1) && ($i < $len2); ++$i) {
    my $c1 = substr($a, $i, 1);
    my $c2 = substr($b, $i, 1);
    ($c1 =~ /^\d/o) || ($c2 =~ /^\d/o) || ($c1 ne $c2) and last;
  }
  my $a_r = ($i < $len1) ? substr($a, $i) : "";
  my $b_r = ($i < $len2) ? substr($b, $i) : "";
  my ($a_n, $a_s) = ($a_r =~ /^(\d+)(.*)$/o);
  my ($b_n, $b_s) = ($b_r =~ /^(\d+)(.*)$/o);
  return (defined($a_n) && defined($b_n)) ?
  (($a_n <=> $b_n) || ($a_s cmp $b_s)) : ($a cmp $b);
}


sub isBLAT($) {
  my ($file) = @_;
  local *FH;

  open(FH, "< $file") or die("$0: Cannot open input file $file:  $!");
  my $line = <FH>;
  close FH;
  if ($line =~ /^.+\s\d+\s\d+;\s.+\s\d+\s\d+;\sscore/) {
    return 0;
  } elsif ($line =~ /^\S+\s\d+\s\d+\s\S+\s/) {
    return 1;
  } else {
    die("$0: Unknown input format in $file. Stopped");
  }
}


sub getMinSeqScore($) {
  my ($file) = @_;
  my $score; local *FH;
  open(FH, "< ".$file) or die("$0: Could not open SLAGAN scorefile $file: $!");
  while (<FH>) {
    # sample line: {+U+;+U-;-U+;-U-}{70000 0 0 0}
    /\{\+U\+\;.+\}.*\{(\d+)\s.+\}/;
    $score = $1 if $1;
  }
  close FH;
  die("$0: Could not determine min_seq_score from SLAGAN scorefile $file. Stopped") unless $score;
  print "$0: min_seq_score: $score\n" if $debug;
  return $score;
}


sub writeSizes($$) {
  my ($sizes, $outfile) = @_;
  local *FH;

  open(FH, "> $outfile")
    or die("$0: Could not open file $outfile for writing: $!");
  foreach my $key (sort alnum keys %$sizes) {
    print FH "$key\t$$sizes{$key}\n";
  }
  close FH;
}


# Borrowed from if.pm to enable standalone conditional module loading on earlier versions of Perl
sub useIf($$) {
  my $method = 'import';
  return unless shift; # condition

  my $package = $_[0];
  (my $file = $package.".pm") =~ s!::!/!g;
  require $file;
  my $method_entry_point = $package->can($method);
  goto &$method_entry_point if $method_entry_point;
}


sub checkAlignCoords($) {
  my ($cur_align) = @_;

  if ($$cur_align[START1] > $$cur_align[END1]) {
    ($$cur_align[START1], $$cur_align[END1]) =
      ($$cur_align[END1], $$cur_align[START1]);
  }
  if ($$cur_align[START2] > $$cur_align[END2]) {
    ($$cur_align[START2], $$cur_align[END2]) =
      ($$cur_align[END2], $$cur_align[START2]);
  }
#  if ($$cur_align[OSTART1] > $$cur_align[OEND1]) { my $i = $$cur_align[OSTART1]; $$cur_align[OSTART1] = $$cur_align[OEND1]; $$cur_align[OEND1] = $i; }
#  if ($$cur_align[OSTART2] > $$cur_align[OEND2]) { my $i = $$cur_align[OSTART2]; $$cur_align[OSTART2] = $$cur_align[OEND2]; $$cur_align[OEND2] = $i; }
}


=head1 NAME

Supermap: Piecewise monotonic alignment map generator for shuffle-lagan

=head1 SYNOPSIS

supermap.pl (gen2=id | sizes2=filename) (gen1=id | sizes1=filename)
[-infile=<file>] [-outfile=<file>] [-bacteria] [-score=filename] [-f]
[file1 file2 ...]

=head1 EXAMPLES

supermap.pl -sizes1=human.sizes -sizes2=mouse.sizes hm.chr*.chaos

=head1 DESCRIPTION

Supermap is a whole-genome alignment map generator. It is an extension to the
Shuffle-LAGAN suite (Brudno et al., 2003). Supermap removes the asymmetry between
the query genomes by running multiple SLAGAN passes and combining them into a full
two-genome alignment.

To run Supermap without the Berkeley Genome Pipeline functionality, you will need
sequence length files for each of the genomes. Each file should contain one sequence
length entry per line, of the form "sequence_name sequence_length".

In the CHAOS output format (this program's input), negative orientation always means second pair of coords is inverted.
In this program's output, negative orientation does not invert coordinates (coordinate pairs are always ascending).

Run supermap.pl with no arguments to see a further description.

The terms "hit" and "anchor" usually refer to local alignments produced by CHAOS or another program.
The term "chain" refers to an extended union of a number of these local alignments.

=head1 DEPENDENCIES

Supermap depends on Utils.pm, SLAGAN, and a number of Unix utilities.

To use the Berkeley Genome Pipeline and cluster functionality, Supermap needs
GPutils.pm, GPDBI.pm, and clust_run.

=head1 LIMITATIONS

Supermap is designed to allow the manipulation of large datasets in a reasonable memory footprint.
To do this, it allows multiple files on input and keeps most of its intermediate data in small temporary files.
However, one current limitation is that the alignments for any sequence in either genome must fit into the largest
addressable file size (typically 2GB), and the output alignments must also fit in that size (the remainder will be truncated).

=head1 BUGS

=head1 TODO

TODO: bacteria description, examples, other input formats
TODO: discuss input glob parameters
TODO: local multithreading
TODO: ignore escaped slashes when splitting dir/file (copy Alex)
TODO: check for ++ etc in SLAGAN out
TODO: .supermaprc file for score files, etc
TODO: hazelton.lbl.gov/bugzilla for supermap

=head1 AUTHOR

Andrey Kislyuk L<mailto:kislyuk@ocf.berkeley.edu>.

=cut
