#!/usr/bin/env perl

use strict;
use Getopt::Long;
use File::Copy;
use Utils;

$0 = rindex($0, "/") > -1 ? substr($0, rindex($0, "/")+1) : $0;

my $VERSION = (
  '$Id: rm_chaos_dup,v 1.11 2011/02/03 21:02:33 poliakov Exp $'
  =~ /,v\s+(\d+\S+)/o)[0] || "TEST";

exit(main());

sub main {
  my ($help, $dir, $infile, $outfile, @files, @outfiles, $out, $overlap,
    $merge, $dup);
  my $min_score = 0;
  
  GetOptions(
    "dir=s" => \$dir,
    "reg_exp=s" => \$infile,
    "outfile=s" => \$outfile,
    "overlap=i" => \$overlap,
    "merge=i" => \$merge,
    "dup" => \$dup,
    "min_score=i" => \$min_score);
  $dup || $overlap || $merge or return usage();
  $infile and
    @files = map { ($dir ? "$dir/" : "") . $_ } safe_glob($infile, $dir);
  @ARGV and push(@files, @ARGV);
  @files or push(@files, *STDIN{IO});
  $out = open_outfile($outfile, ref($files[0]), \@outfiles);
  foreach my $file (@files) {
    my $in = ref($file) ? $file : open_infile($file);
    $out = parse_chaos($in, $out, \@outfiles, $overlap, $merge, $dup,
      $min_score);
    ref($file) or close($in);
  }
  @outfiles and close($out), tmp2stdout(\@outfiles);
  return 0;
}

sub usage {
  print STDERR <<EOF;

  $0 (ver. $VERSION) - remove duplicates in CHAOS files

  Usage: $0 [-dir=<dir>] [-reg_exp=<reg_exp>] [-dup] [-min_score=S] \
           [-overlap=O] [-merge=M] [-outfile=<filename>] \
	   [infile1 [infile2] ...]
EOF
  return 1;
}

sub open_infile {
  my ($infile) = @_;
  my ($fh);
  
  if (defined($infile)) {
    $fh = open_file($infile, "<") or die $Utils::Error;
  }
  return $fh;
}

sub open_outfile {
  my ($outfile, $stdin, $outfiles) = @_;
  my ($fh);
  my $tmpfile = "/tmp/$0.$$.1";
  
  if (defined($outfile)) {
    $fh = open_file($outfile, ">") or die $Utils::Error;
    return $fh;
  }
  $stdin or return *STDOUT;
  push(@$outfiles, $tmpfile);
  $fh = open_file($tmpfile, ">") or die $Utils::Error;
  return $fh;
}

sub parse_chaos {
  my ($in, $out, $outfiles, $overlap, $merge, $dup, $min_score) = @_;
  my ($size);
  my $re = '^(\S+)\s+(\d+)\s+(\d+);\s+(\S+)\s+(\d+)\s+(\d+);\s+' .
    'score\s+=\s+(\d+)(?:\.\d+)?\s+\(([-\+])\)';
  my (@saved_hits);
  my $max_end = 0;

  while (<$in>) {
    my ($name1, $start1, $end1, $name2, $start2, $end2,
      $score, $strand) = ($_ =~ /$re/o);
    $strand or next;
    $start2 > $end2 and ($start2, $end2) = ($end2, $start2);
    if ($max_end &&
        (($name1 ne $saved_hits[0][0]) || ($start1 > $max_end))) {
      $merge and merge_dup(\@saved_hits, $merge);
      $overlap || $dup and remove_dup(\@saved_hits, $overlap, $dup, $min_score);
      @saved_hits and
        ($out, $size) = print_hits(\@saved_hits, $out, $outfiles, $size),
	@saved_hits = ();
      $max_end = 0;
    }
    push(@saved_hits, [$name1, $start1, $end1, $name2, $start2, $end2,
      $score, $strand]);
    $end1 > $max_end and $max_end = $end1;
  }
  if (@saved_hits) {
      $merge and merge_dup(\@saved_hits, $merge);
      $overlap || $dup and remove_dup(\@saved_hits, $overlap, $dup, $min_score);
  }
  @saved_hits and
    ($out, $size) = print_hits(\@saved_hits, $out, $outfiles, $size);
  return $out;
}

sub merge_dup {
  my ($hits, $merge) = @_;
  
  for (my $i = 0; $i < $#$hits; $i++) {
    my ($name1i, $start1i, $end1i, $name2i, $start2i, $end2i, $scorei,
      $strandi) = @{$$hits[$i]};
    my $len1 = $end1i - $start1i + 1;
    my $len2 = $end2i - $start2i + 1;
    for (my $j = $i + 1; $j < @$hits; $j++) {
      my ($name1j, $start1j, $end1j, $name2j, $start2j, $end2j, $scorej,
        $strandj) = @{$$hits[$j]};
      ($start1j > $end1i) and last;
      ($name2j ne $name2i) || ($strandj ne $strandi) ||
          ($start2j > $end2i) || ($end2j < $start2i) and
	next;
      my $i_over_u1 = (Min($end1i, $end1j) - $start1j + 1) /
	(Max($end1i, $end1j) - $start1i + 1);
      (100 * $i_over_u1) < $merge and next;
      my $i_over_u2 = (Min($end2i, $end2j) - Max($start2i, $start2j) + 1) /
        (Max($end2i, $end2j) - Min($start2i, $start2j) + 1);
      (100 * $i_over_u2) < $merge and next;
      my $merged_score = Round(($scorei + $scorej) *
        ($i_over_u1 + $i_over_u2) / 2);
      $merged_score < Max($scorei, $scorej) and next;
      $end1i < $end1j and $end1i = $end1j;
      $start2i > $start2j and $start2i = $start2j;
      $end2i < $end2j and $end2i = $end2j;
      $scorei = $merged_score;
      @{$$hits[$i]}[2,4..6] = ($end1i, $start2i, $end2i, $merged_score);
      splice(@$hits, $j, 1);
      $j--, next;
    }
  }
}

sub remove_dup {
  my ($hits, $overlap, $dup, $min_score) = @_;
  
  for (my $i = 0; $i < $#$hits; $i++) {
    my ($name1i, $start1i, $end1i, $name2i, $start2i, $end2i, $scorei,
      $strandi) = @{$$hits[$i]};
    my ($chk_dup);
    $dup and $chk_dup = (($end1i - $start1i) == ($end2i - $start2i));
    $overlap || $chk_dup or next;
    my $len1 = $end1i - $start1i + 1;
    my $len2 = $end2i - $start2i + 1;
    for (my $j = $i + 1; $j < @$hits; $j++) {
      my ($name1j, $start1j, $end1j, $name2j, $start2j, $end2j, $scorej,
        $strandj) = @{$$hits[$j]};
      ($start1j > $end1i) and last;
      ($name2j ne $name2i) || ($strandj ne $strandi) and next;
      if ($overlap) {
        ($start2j > $end2i) || ($end2j < $start2i) and next;
	if (((100 * (Min($end1i, $end1j) - $start1j + 1) /
	    Min($len1, $end1j - $start1j + 1)) >= $overlap) &&
            ((100 * (Min($end2i, $end2j) - Max($start2i, $start2j) + 1) /
	    Min($len2, $end2j - $start2j + 1)) >= $overlap)) {
          my $next_j = ($scorei >= $scorej);
          splice(@$hits, $next_j ? $j : $i, 1);
          $next_j and $j--, next;
          $i--;
          last;
	}
      }
      $chk_dup or next;
      (($end1j - $start1j) != ($end2j - $start2j)) ||
	  (($start1j - $start1i) != (($strandi eq "+") ?
	  ($start2j - $start2i) : ($end2i - $end2j))) and
	next;
      $scorei = Round($scorei * ($start1j - $start1i) / $len1);
      $end1i = $start1j - 1;
      if ($strandi eq "+") {
        $end2i = $start2j - 1;
      } else {
        $start2i = $end2j + 1;
      }
      $len1 = $end1i - $start1i + 1;
      $len2 = $end2i - $start2i + 1;
      ($scorei < $min_score) || ($len1 < 2) || ($len2 < 2) and
        splice(@$hits, $i, 1), $i--, last;
      @{$$hits[$i]}[2,4..6] = ($end1i, $start2i, $end2i, $scorei);
    }
  }
}

sub print_hits {
  my ($hits, $out, $outfiles, $size) = @_;
  
  if (@$outfiles) {
    $size > 2000000000 and
      $out = switch_file($out, $outfiles),
      $size = 0;
  }
  foreach my $hit (@$hits) {
    my $line = "$$hit[0] $$hit[1] $$hit[2]; $$hit[3] " .
      (($$hit[7] eq "+") ? "$$hit[4] $$hit[5]" : "$$hit[5] $$hit[4]") .
      "; score = $$hit[6].0 ($$hit[7])\n";
    @$outfiles and $size += length($line);
    print $out $line;
  }
  return ($out, $size);
}

sub switch_file {
  my ($out, $outfiles) = @_;
  my ($file, $fh);

  close($out);
  $file = $$outfiles[$#$outfiles];
  $file =~ s/\d+$/$& + 1/e;
  push(@$outfiles, $file);
  $fh = open_file($file, ">") or die $Utils::Error;
  return $fh;
}

sub tmp2stdout {
  my ($outfiles) = @_;

  foreach my $file (@$outfiles) {
    copy($file, \*STDOUT);
    unlink($file);
  }
}
